#!/usr/bin/python3
# -*- coding: utf8 -*-

# Copyright (c) 2021 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
Quanlse's Scheduler is a module that allows for automatic generation of fidelity-optimized and
scheduled pulse sequence for a given quantum circuit set to perform a certain quantum computing task.

Quanlse's Scheduler has the following benefits:

- Highly automatic: it generates high-fidelity pulses automatically and simultaneously
  minimizes the overall gate operation time.

- Practical: it considers several limitations of the superconducting quantum system,
  including leakage errors, decoherence, etc.

- Flexible: it gives users the freedom to customize qubits and circuit parameters.
  This can also be easily extended to other quantum computing platforms.

At present, Quanlse Scheduler support two different quantum platforms:

**1. Superconduct** (:doc:`Quanlse.Scheduler.Superconduct`)

  - We provide :doc:`Quanlse.Scheduler.Superconduct.DefaultPulseGenerator` using :doc:`remoteOptimizer`
    to users to generate optimal pulses by accessing the Quanlse Cloud Service.

  - We provide :doc:`Quanlse.Scheduler.Superconduct.RBPulseGenerator` using :doc:`remoteOptimizer` especially
    for the :doc:`Quanlse.Utils.RandomizedBenchmarking` functions.

  - We provide :doc:`Quanlse.Scheduler.Superconduct.DefaultPipeline` to users to support the fundamental scheduling
    strategy.

**2. Ion Trap** (:doc:`Quanlse.Scheduler.Ion`)

For more details, please visit https://quanlse.baidu.com/#/doc/tutorial-scheduler.
"""

import copy
from typing import List, Union, Optional, Dict, Any

import numpy

from Quanlse.QHamiltonian import QHamiltonian as QHam
from Quanlse.QOperation.FixedGate import FixedGateOP
from Quanlse.QOperation.RotationGate import RotationGateOP
from Quanlse.QWaveform import QJob
from Quanlse.QRegPool import QRegPool
from Quanlse.QOperation import Error, CircuitLine
from Quanlse.Scheduler.SchedulerPipeline import SchedulerPipeline
from Quanlse.Scheduler.SchedulerPulseGenerator import SchedulerPulseGenerator
from Quanlse.Scheduler.GatePulsePair import GatePulsePair


class Scheduler:
    """
    Basic class of Quanlse Scheduler.

    :param dt: AWG sampling time
    :param ham: the QHamiltonian object.
    :param pulseGenerator: the pulseGenerator object.
    :param subSysNum: size of the subsystem
    :param sysLevel: the energy levels of the system (support different levels for different qubits)
    :param savePulse: save the pulse in cache
    """

    def __init__(self, dt: float = None, ham: QHam = None, generator: SchedulerPulseGenerator = None,
                 pipeline: SchedulerPipeline = None, subSysNum: int = None, sysLevel: int = None,
                 savePulse: bool = True):
        """
        Initialization
        """

        if sysLevel is not None:
            self._sysLevel = sysLevel  # type: Union[int, List[int]]
        else:
            self._sysLevel = 2  # type: Union[int, List[int]]

        if subSysNum is not None:
            self._subSysNum = subSysNum  # type: int
        else:
            self._subSysNum = 1  # type: int

        # The quantum register
        self.Q = QRegPool(self)  # type: QRegPool

        # The list for saving the quantum circuit
        self._circuit = []  # type: List['CircuitLine']

        # The configuration dictionary
        self._conf = {}  # type: Dict[str, Any]

        # Sampling duration of AWG (arbitrary wave generator)
        self._dt = dt  # type: float

        # The vault for store the pulses of the fixed gates
        self._pulseCache = []  # type: List[GatePulsePair]

        # The time padding of the pulses of adjacent quantum gates
        self.padding = 0.0  # type: float

        # Save the control operators
        self._tempCtrlOperators = {}  # type: Dict[str, Any]

        # The QHam object, this is not necessary
        self._ham = None  # type: Optional[QHam]
        if ham is not None:
            self._ham = copy.deepcopy(ham)
            self._ham.dt = dt
            self._subSysNum = ham.subSysNum
            self._sysLevel = ham.sysLevel

        # Add the pulse generator
        if generator is not None:
            self._pulseGenerator = generator  # type: Optional[SchedulerPulseGenerator]
        else:
            self._pulseGenerator = None  # type: Optional[SchedulerPulseGenerator]

        # If save the pulse for quantum gates in cache
        self._savePulse = savePulse  # type: bool

        # Add the pipeline
        if pipeline is not None:
            self._pipeline = pipeline  # type: SchedulerPipeline
        else:
            self._pipeline = SchedulerPipeline()  # type: SchedulerPipeline

        # The QJob object, this is not necessary
        self._initializeQJob(dt, subSysNum, sysLevel)

    @property
    def conf(self) -> Dict[str, Any]:
        """
        Scheduler Configuration
        """
        return self._conf

    @property
    def pipeline(self) -> SchedulerPipeline:
        """
        Scheduler Pipeline
        """
        return self._pipeline

    @pipeline.setter
    def pipeline(self, pipeline: SchedulerPipeline):
        """
        Pipeline setter
        """
        if self._pipeline is not None:
            del self._pipeline
        self._pipeline = pipeline

    @property
    def savePulse(self) -> bool:
        """
        If save the pulse for quantum gates in cache
        """
        return self._savePulse

    @savePulse.setter
    def savePulse(self, value: bool):
        """
        If save the pulse for quantum gates in cache
        """
        self._savePulse = value

    @property
    def subSysNum(self) -> int:
        """
        Return the number of sub-systems.
        """
        return self._subSysNum

    @property
    def sysLevel(self) -> Union[int, List[int]]:
        """
        Return the size of the sub-systems.
        """
        return self._sysLevel

    @property
    def dt(self) -> float:
        """
        AWG Sampling time
        """
        return self._dt

    @property
    def circuit(self) -> List[CircuitLine]:
        """
        Circuit information
        """
        return self._circuit

    @circuit.setter
    def circuit(self, circuit: List[CircuitLine]):
        """
        Circuit setter
        """
        self._circuit = circuit
        self._initializeQJob(self._dt, self._subSysNum, self._sysLevel)

    @property
    def pulseCache(self) -> List[GatePulsePair]:
        """
        Pulse cache
        """
        return self._pulseCache

    @property
    def pulseGenerator(self) -> Optional[SchedulerPulseGenerator]:
        """
        Pulse cache
        """
        return self._pulseGenerator

    @pulseCache.setter
    def pulseCache(self, obj: List[GatePulsePair]):
        """
        Pulse cache setter
        """
        self._pulseCache = copy.deepcopy(obj)

    @property
    def ham(self) -> Optional[QHam]:
        """
        Respective Hamiltonian
        """
        if self._ham is None:
            return None
        else:
            return self._ham

    @ham.setter
    def ham(self, obj: QHam):
        """
        Hamiltonian setter
        """
        self._ham = copy.deepcopy(obj)
        self._subSysNum = obj.subSysNum
        self._sysLevel = obj.sysLevel

    @property
    def job(self) -> QJob:
        """
        QJob in Scheduler
        """
        return self._job

    def _initializeQJob(self, dt, subSysNum: int = None, sysLevel: int = None) -> None:
        """
        Initialize a QJob object.

        :param dt: AWG sampling time
        :param subSysNum: sybsystem's number
        :param sysLevel: system's level
        :return: None
        """
        if self._ham is None:
            if subSysNum is None:
                for cirLin in self._circuit:
                    if self._subSysNum < cirLin.qRegIndexList.max() + 1:
                        self._subSysNum = cirLin.qRegIndexList.max() + 1
            else:
                self._subSysNum = subSysNum

            if sysLevel is None:
                raise Error.ArgumentError("You must input sysLevel when ham is None!")
            else:
                self._sys = sysLevel
            self._job = QJob(self._subSysNum, self._sysLevel, dt)
        else:
            self._tempCtrlOperators = copy.deepcopy(self._ham.job.ctrlOperators)
            self._ham.job = QJob(self.ham.subSysNum, self.ham.sysLevel, self.ham.dt)
            self._ham.job.ctrlOperators = copy.deepcopy(self._tempCtrlOperators)

    def setGatePulsePair(self, *args) -> None:
        """
        Add a GatePulsePair instance to the pulse cache.

        :return: None
        """
        if isinstance(args, GatePulsePair):
            # User input a GatePulsePair object directly
            foundGate = self.findGatePulsePair(args.cirLine.data)
            if foundGate is not None:
                foundGate.pulses = copy.deepcopy(args.pulses)
            else:
                pair = copy.deepcopy(args)
                self._pulseCache.append(pair)
        elif isinstance(args, tuple):
            # User input a gate and a QJob
            if isinstance(args[0].data, (FixedGateOP, RotationGateOP)) and \
                    (isinstance(args[1], QJob) or args[1] is None):
                foundGate = self.findGatePulsePair(args[0])
                if foundGate is not None:
                    foundGate.pulses = copy.deepcopy(args[1])
                else:
                    pair = GatePulsePair(args[0], args[1])
                    self._pulseCache.append(pair)
            else:
                raise Error.ArgumentError(f"Wrong argument, please input a GatePulsePair "
                                          f"or a pair of FixedGate and QJob objects. But you input "
                                          f"a pair of {type(args[0])} with {type(args[1])}.")

    def findGatePulsePair(self, cirLine: CircuitLine) -> Union[GatePulsePair, None]:
        """
        Find the pulse of a fixed gate in pulse cache.

        :param cirLine: CircuitLine object to be searched in pulse cache
        :return: the returned GatePulsePair
        """
        for pair in self._pulseCache:
            if GatePulsePair(cirLine) == pair:
                return pair
        return None

    def getMatrix(self) -> numpy.ndarray:
        """
        Return the full matrix of circuit.
        """
        dim = 2 ** self.subSysNum
        # Initialize the matrix
        cirMat = numpy.identity(dim)
        # Traverse the circuitLine
        for cir in self.circuit:
            # Obtain basic data
            _mat = cir.data.getMatrix()
            _onQubits = cir.qRegIndexList
            if max(_onQubits) >= self.subSysNum:
                raise Error.ArgumentError("Qubit index(es) exceeds the scheduler's subSysNum.")
            if 1 <= len(_onQubits) <= 2:
                if _mat.shape[0] != 2 ** len(_onQubits):
                    raise Error.ArgumentError(f"The matrix of {len(_onQubits)}-qubit gate must be {2 ** len(_onQubits)}"
                                              f"-dimensional, however its dimension is {_mat.shape[0]}.")
            else:
                raise Error.ArgumentError("Only single- or two-qubit gates are supported!")
            if len(_onQubits) > 1 and abs(_onQubits[0] - _onQubits[1]) > 1:
                raise Error.ArgumentError("The indexes of two qubit gates must be close.")
            # Generate the matrix
            _preIdm = numpy.identity(2 ** min(_onQubits))
            _postIdm = numpy.identity(2 ** (self.subSysNum - max(_onQubits) - 1))
            cirMat = numpy.kron(numpy.kron(_preIdm, _mat), _postIdm) @ cirMat

        return cirMat

    def plot(self) -> None:
        """
        Plot the pulses.

        :return: None
        """
        if self._ham is not None:
            self._ham.job.plot()
        elif self._job is not None:
            self._job.plot()

    def plotIon(self) -> None:
        """
        Plot the ion pulses.

        :return: None
        """
        if self._ham is not None:
            self._ham.job.plotIon()
        elif self._job is not None:
            self._job.plotIon()

    def clearCircuit(self) -> None:
        """
        Clear all circuit.

        :return: None
        """
        self._circuit = []
        if self._ham is not None:
            self._ham.job.clearWaves()
        elif self._job is not None:
            self._job.clearWaves()

    def clearCache(self) -> None:
        """
        Clear all cache.

        :return: None
        """
        self._pulseCache = []

    def schedule(self) -> QJob:
        """
        Run the scheduler.

        :return: a scheduled QJob object
        """
        # Step 1: generate pulses for the GatePulsePair instances
        # When self._ham is None, it means that the Scheduler does not need to
        # generate pulses for the CircuitLine instances.
        if self._pulseGenerator is not None:
            if self.savePulse:
                for cirLine in self.circuit:
                    foundPair = self.findGatePulsePair(cirLine)
                    if foundPair is None or foundPair.pulses is None:
                        pulseJob = self._pulseGenerator(cirLine, self)
                        self.setGatePulsePair(cirLine, pulseJob)
        else:
            raise Error.ArgumentError("No pulseGenerator is set, hence cannot generate pulse.")

        # Obtain a copy of job
        _tmpJob = copy.deepcopy(self._job) if self._ham is None else copy.deepcopy(self._ham.job)

        # Step 2: Assemble all the pulses according to self.circuit
        _tmpJob.clearWaves()
        for cirLine in self._circuit:
            if self.savePulse:
                _pulses = self.findGatePulsePair(cirLine).pulses
            else:
                _pulses = self._pulseGenerator(cirLine, self)
            _tmpJob.appendJob(_pulses)

        # Step 3: Run the scheduling process
        _tmpJob = self.pipeline(job=_tmpJob, scheduler=self)
        return _tmpJob



#!/usr/bin/python3
# -*- coding: utf8 -*-

# Copyright (c) 2021 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Pulse RPC
"""
import os
from pprint import pprint
from enum import Enum
import json

from Quanlse.Define import sdkVersion, taskSourceQuanlse, circuitPackageFile, outputPath
from Quanlse.Define.Settings import outputInfo
from Quanlse.QPlatform import Error

from Quanlse.QTask import (
    invokeBackend,
    _retryWhileNetworkError,
    _uploadCircuit,
    _waitTask,
)

OPTIMIZER_TO_TASKTYPE = {
    "runHamiltonian": "quanlse_simulator",
    "zneMitigation": "quanlse_zne",
}


@_retryWhileNetworkError
def _createTask(token, circuitId, optimizer, backendParam=None, modules=[], debug=False, taskType="quanlse_optimizer"):
    """
    Create a task from the code
    """

    task = {
        "token": token,
        "circuitId": circuitId,
        "taskType": taskType,
        "sdkVersion": sdkVersion,
        "source": taskSourceQuanlse,
        "optimizer": optimizer,
        "modules": modules
    }

    if debug:
        task['debug'] = debug

    if backendParam:
        paramList = []
        for param in backendParam:
            if isinstance(param, Enum):
                paramList.append(param.value)
            else:
                paramList.append(param)
        task['backendParam'] = paramList

    ret = invokeBackend(
        "task/createTask",
        task
    )

    return ret['taskId']


def rpcCall(optimizer, args, kwargs, debug=False):
    """
    invoke remote optimizer
    """
    circuitId = None
    taskId = None

    program = {
        "optimizer": optimizer,
        "args": args,
        "kwargs": kwargs,
    }

    programBuf = json.dumps(program)
    with open(circuitPackageFile, 'wb') as file:
        file.write(programBuf.encode("utf-8"))

    token, circuitId = _uploadCircuit(circuitPackageFile)
    taskType = OPTIMIZER_TO_TASKTYPE.get(optimizer, "quanlse_optimizer")
    taskId = _createTask(token, circuitId, optimizer, debug, taskType=taskType)

    if outputInfo:
        print(f"Circuit uploaded successfully, circuitId => {circuitId} taskId => {taskId}")

    taskResult = _waitTask(token, taskId, downloadResult=True)

    if type(taskResult) == str:
        raise Error.RuntimeError(taskResult)

    errorMsg = taskResult.get("outer_error")
    if errorMsg:
        raise Error.RuntimeError(errorMsg)

    else:
        localFile = os.path.join(outputPath, f'remote.{taskId}.origin.json')
        with open(localFile, "rb") as fObj:
            origin = json.loads(fObj.read())
            # return createFromJson(origin["qam"]), origin["infidelity"]
            return origin



#!/usr/bin/python3
# -*- coding: utf8 -*-

# Copyright (c) 2021 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
In a superconducting quantum system, the control Hamiltonian represents external
control (microwave, magnetic flux etc.) acting on the qubits.

In Quanlse, users can use function ``addWave()`` or ``appendWave()`` to define and add wave to system.
Additionally, users can use ``clearWaves()`` to clear all waveforms in the specified control term.

Here, we will take ``addWave()`` or ``appendWave()`` to define and add the control wave. Each waveform
function ``A(t)`` can be defined using four arguments: start time ``t0``, duration ``t`` and the
corresponding parameters ``a``, ``tau``, ``sigma``. The function ``addWave()`` or ``appendWave()``
allows us to define waveforms in different ways:

**Example 1: using ``addWave()`` to add preset waveform functions**

Users can use the preset waveforms, here we use the Gaussian wave ``gaussian`` as an example:

.. code-block:: python

        ham.addWave(driveX, onSubSys=0, waves=gaussian(t=20, a=1.1, tau=10, sigma=4), t0=0)

**Example 2: using ``appendWave()`` to append preset waveform functions to the end of the sequence**

Users can use the preset waveforms, here we use the Gaussian wave ``gaussian`` as an example. The following
codes can append the waveform to the end of the existing sequence of the current control term `driveX`:

.. code-block:: python

        ham.appendWave(driveX, onSubSys=0, waves=gaussian(t=20, a=1.1, tau=10, sigma=4))

Users also can set ``compact`` parameter to ``False`` to append the waveform into the end of the whole sequence
of all control terms:

.. code-block:: python

        ham.appendWave(driveX, onSubSys=0, waves=gaussian(t=20, a=1.1, tau=10, sigma=4), compact=False)

**Example 3: using user-defined wave functions**

Users can also define a function in the form of ``func(_t, args)``, where the first parameter ``_t`` is the time
duration and ``args`` is the pulse parameters.

.. code-block:: python

        def userWaveform(t: Union[int, float], a: float, tau: float, sigma: float,
                         freq: float = None, phase: float = None) -> QWaveform:
            def func(_t, args):
                _a, _tau, _sigma = args
                if _sigma == 0:
                    return 0
                pulse = _a * exp(- ((_t - _tau) ** 2 / (2 * _sigma ** 2)))
                return pulse

            wave = QWaveform(f=func, t=t, args=(a, tau, sigma), freq=freq, phase=phase)
            wave.name = "user-defined wave"
            return wave
        ham.addWave(driveX, onSubSys=0, waves=userWaveform(t=20, a=1.1, tau=10, sigma=4), t0=0)


For more details, please visit https://quanlse.baidu.com/#/doc/tutorial-construct-ham.
"""

import copy
import math
import pickle
import base64
import numpy

from math import floor, sqrt, pi, atan
from scipy.special import erf
import matplotlib.pyplot as plt
from collections.abc import KeysView
from numpy import array, shape, linspace, arange, exp, sign
from typing import Dict, Any, List, Union, Tuple, Callable, Optional

from Quanlse.QPlatform import Error
from Quanlse.QOperator import QOperator, driveX, driveY
from Quanlse.Utils.Plot import plotPulse
from Quanlse.Utils.Functions import generateOperatorKey, combineOperatorAndOnSubSys, formatOperatorInput

QWAVEFORM_FLAG_ALWAYS_ON = 0b00001  # type: int
"""
QWaveform Flag: Users can input this flag to ``QJob.addWave()`` or ``QJob.appendWave()`` method to 
realize following function:

The waveform added will ignore the provided gate time, and adjust the gate time
according to the ``maxEndTime`` of the QJob.
"""

QWAVEFORM_FLAG_DO_NOT_CLEAR = 0b00010  # type: int
"""
QWaveform Flag: Users can input this flag to ``QJob.addWave()`` or ``QJob.appendWave()`` method to 
realize following function:

The waveform added will never be cleared by ``QJob.clearWaves()`` or ``QJob.appendWave()`` method,
and will be saved in a QHamiltonian instance, instead of a QJob instance.
Hence, when the QJob instance has no parent instance (QHamiltonian), no
QWAVEFORM_FLAG_DO_NOT_CLEAR waveform can be added.
"""

QWAVEFORM_FLAG_DO_NOT_PLOT = 0b00100  # type: int
"""
QWaveform Flag: Users can input this flag to ``QJob.addWave()`` or ``QJob.appendWave()`` method to 
realize following function:

The waveform added will be ignored by ``QJob.plot()``.
"""

QWAVEFORM_FLAG_PHASE_SHIFT = 0b01000  # type: int
"""
QWaveform Flag: Users can input this flag to ``QJob.addWave()`` or ``QJob.appendWave()`` method to 
realize following function:

The waveform cause some amount of phase shift. The amplitude will be ignore.
"""

QWAVEFORM_FLAG_MIX_WAVE = 0b10000  # type: int
"""
QWaveform Flag: Users can input this flag to ``QJob.addWave()`` or ``QJob.appendWave()`` method to 
realize following function:

This kind of waveform mix the two waves following the rules of microwave signal.
"""


class QWaveform:
    """
    Class of QWaveform.
    In Quanlse, we use a QWaveform object to store all information regarding a wave. A QWaveform object includes the
    wave's function (Gaussian, Square, QuasiSquare, etc.) and the respective parameters,
    a wave's starting time and duration, as well as the pulse frequency and phase, AWG sampling time, etc. QWaveform
    objects can be added to a control term in the Hamiltonian, giving a control term a specified waveform.

    :param f: a callable function, indicating a specific waveform
    :param t0: wave starting time
    :param t: wave duration
    :param args: wave arguments (takes different parameters for different waveform)
    :param seq: users can also input a python list consisting the wave sequence
    :param dt: AWG sampling time
    :param strength: wave strength
    :param freq: pulse frequency shift
    :param phase: pulse phase shift (will accumulate during the entire pulse execution)
    :param phase0: pulse phase shift (will not accumulate during the entire pulse execution)
    :param tag: wave tag indicating purpose
    :param flag: wave flag
    """

    def __init__(self, f: Callable = None, t0: Union[int, float] = 0.0, t: Union[int, float] = 0.0,
                 args: Any = None, seq: List[float] = None, dt: float = None, strength: float = 1.,
                 freq: float = None, phase: float = None, phase0: float = None, tag: str = None, flag: int = 0) -> None:
        """
        Constructor for QWaveform class.
        """

        if seq is not None and f is not None:
            raise Error.ArgumentError("Cannot input seq and f at the same time.")

        if f is None:
            if seq is None:
                raise Error.ArgumentError("You should input either func or seq.")
            # Record necessary information of the wave
            self.func = None  # type: Optional[Union[Callable, str]]
            self.t = None  # type: Optional[float]
            self.seq = copy.deepcopy(seq)  # type: Optional[List[Union[float, complex]]]
            self.args = None  # type: Optional[Any]
        elif callable(f) or isinstance(f, str):
            if seq is not None:
                print("WARNING: func is given, hence the input of seq is ignored!")
            # Record necessary information of the wave
            self.func = f  # type: Optional[Union[Callable, str]]
            self.t = t  # type: Optional[float]
            self.seq = None  # type: Optional[List[Union[float, complex]]]
            self.args = copy.deepcopy(args)  # type: Optional[Any]
        else:
            raise Error.ArgumentError("Unsupported type of input for func, it should either"
                                      " be a string, function or None.")

        # Common arguments.
        self.t0 = float(t0)  # type: float
        self.name = ""  # type: Optional[str]
        self.dt = None if dt is None else float(dt)  # type: Optional[float]
        self.strength = float(strength)  # type: float
        self.freq = freq  # type: Optional[float]
        self.phase = phase  # type: Optional[float]
        self.phase0 = phase0  # type: Optional[float]
        self.tag = tag  # type: Optional[str]
        self.flag = flag  # type: int

    def __str__(self) -> str:
        """
        Print description of the object.
        """
        returnStr = ""
        returnStr += f"Waveform: {self.name}\n"
        returnStr += f"    - tag: {self.tag}\n"
        returnStr += f"    - t0: {self.t0}\n"
        returnStr += f"    - t: {self.t}\n"
        returnStr += f"    - seq: {None if self.seq is None else len(self.seq)}\n"
        returnStr += f"    - args: {self.args}\n"
        returnStr += f"    - freq: {self.freq}\n"
        returnStr += f"    - phase: {self.phase}\n"
        returnStr += f"    - drive strength: {self.strength}\n"
        returnStr += f"    - flag: {self.flag}\n"

        return returnStr

    def __call__(self, *args, **kwargs):
        """
        Call self.func directly.
        """
        t = args[0]
        if self.func is not None:
            if len(args) == 1:
                return self.func(t, self.args)
            elif len(args) == 2:
                return self.func(t, args)
            elif len(args) == 0:
                raise Error.ArgumentError("You should input the time argument!")
            else:
                raise Error.ArgumentError(f"Too many arguments ({len(args)})!")
        elif self.seq is not None:
            if self.dt is None:
                raise Error.ArgumentError("Property dt is not set, cannot used to call a sequence-type QWaveform.")
            realTime = t
            if int(realTime / self.dt) >= len(self.seq):
                return 0.0
            else:
                return self.seq[int(realTime / self.dt)]
        else:
            raise Error.ArgumentError("Neither sequence nor func is set.")

    def waveFunctionToSequence(self, dt: float, maxEndTime: float = None) -> 'QWaveform':
        """
        Transform the function in QWaveform object to a sequence.

        :param dt: AWG sampling time
        :param maxEndTime: maximum ending time
        :return: returned QWaveform object
        """

        if self.func is not None:
            if dt is not None:
                _dt = dt
            elif self.dt is not None:
                _dt = self.dt
            else:
                raise Error.ArgumentError("Argument dt is not specified, please input dt.")
            seqList = []
            # QWAVEFORM_FLAG_ALWAYS_ON flag
            if self.flag & QWAVEFORM_FLAG_ALWAYS_ON:
                if maxEndTime is None:
                    print("Warning: QWaveform object with QWAVEFORM_FLAG_ALWAYS_ON flag"
                          " will become useless after serialization. You can input maxE"
                          "ndTime to serialize this object.")
                    endTimeDt = 0
                else:
                    endTimeDt = int(maxEndTime / _dt)
            else:
                endTimeDt = int(self.t / _dt)
            # Generate sequence
            for nowDt in range(endTimeDt):
                nowNs = nowDt * dt + _dt / 2
                seqList.append(self(nowNs))
            wave = sequence(t0=self.t0, seq=seqList, freq=self.freq, phase=self.phase)
            wave.tag = self.tag
            wave.flag = self.flag
            wave.dt = _dt
            return wave
        elif self.seq is not None:
            return copy.deepcopy(self)

    def plot(self, dt: float = None, xUnit: str = 'ns', yUnit: str = 'Amp (a.u.)',
             color: Union[str, List[str]] = None, dark: bool = False) -> None:
        """
        Plot the waveform.

        :param dt: AWG sampling time
        :param xUnit: X coordinate unit
        :param yUnit: Y coordinate unit
        :param color: pulse color options
        :param dark: whether activate dark mode
        :return: None
        """
        x = []
        y = []
        yLabel = []
        colors = []
        colorIndex = 0
        seqList = []
        _dt = None
        if self.func is not None:
            if dt is not None:
                _dt = dt
            elif self.dt is not None:
                _dt = self.dt
            elif self.seq is None:
                raise Error.ArgumentError("Argument dt is not specified, please input dt.")
            endTimeDt = int(self.t / _dt)
            for nowDt in range(endTimeDt):
                nowNs = nowDt * _dt + _dt / 2
                seqList.append(self(nowNs))
            tList = linspace(self.t0, self.t0 + self.t, len(seqList))
        else:
            seqList = self.seq
            tList = arange(0, len(seqList)).astype(int)

        y.append(list(seqList))
        x.append(list(tList))
        yLabel.append(yUnit)

        # Whether repetitive colors or all blue
        if color is None:
            colors.append('blue')
        else:
            colors.append(color[colorIndex])
        if self.name is not None:
            names = [self.name]
        else:
            names = ['']
        plotPulse(x, y, xLabel=f'Time ({xUnit})', yLabel=yLabel, title=names, color=colors, dark=dark)
        plt.show()

    def dump(self, dt: float = None) -> str:
        """
        Return base64 encoded string.

        :param dt: AWG sampling time
        :return: returned base64 encoded string
        """
        obj = self.waveFunctionToSequence(dt)
        # For compatibility with previous versions
        if not hasattr(obj, "omega"):
            obj.omega = obj.freq
            obj.phi = 0.
        if not hasattr(obj, "phi"):
            obj.phi = obj.phase0
        byteStr = pickle.dumps(obj)
        base64str = base64.b64encode(byteStr)
        del obj
        return base64str.decode()

    def copy(self) -> 'QWaveform':
        """
        Return the copy of the object
        """
        return copy.deepcopy(self)

    @staticmethod
    def load(base64Str: str) -> 'QWaveform':
        """
        Create object from base64 encoded string.

        :param base64Str: a input base64 encoded string
        :return: returned QWaveform object
        """
        byteStr = base64.b64decode(base64Str.encode())
        obj = pickle.loads(byteStr)  # type: QWaveform
        # For compatibility with previous versions
        if not hasattr(obj, "freq"):
            obj.freq = obj.omega if hasattr(obj, "omega") else None
        if not hasattr(obj, "phase0"):
            obj.phase0 = obj.phi if hasattr(obj, "phi") else None
        if not hasattr(obj, "phase"):
            obj.phase = None
        return obj


class QJob:
    """
    A QJob object in Quanlse contains all the information regarding a quantum task.
    The corresponding pulse information is stored in waveCache.

    :param subSysNum: subsystem's number
    :param sysLevel: system's energy level
    :param dt: AWG sampling time
    :param title: a user-given title
    :param description: a user-given description, preferably indicating the purpose
    :param ampGamma: An optional parameter indicating the amplitude's Gamma value
    """

    def __init__(self, subSysNum: int, sysLevel: Union[int, List[int]], dt: float,
                 title: str = "", description: str = "", ampGamma: Optional[Union[int, float]] = None) -> None:
        """
        Constructor for QJob class.
        """

        self.title = title  # type: str
        self.description = description  # type: str
        self.dt = dt  # type: float
        self.subSysNum = subSysNum  # type: int
        self.sysLevel = sysLevel  # type: int
        self.tagWhiteListForClearWaves = []

        self.endTimeDt = 0  # type: int
        self.endTime = 0.0  # type: float

        self.ampGamma = ampGamma  # type: Optional[Union[int, float]]

        # Control operators
        self._ctrlOperators = {}  # type: Dict[str, Union[QOperator, List[QOperator]]]
        # List of QWaveform
        self._waves = {}  # type: Dict[str, Any]
        # List of local oscillator
        self._LO = {}  # type: Dict[str, Tuple[float, float]]
        # Sequence of cache for QWaveform
        self._waveCache = {}  # type: Dict[str, Any]
        # Parent object (QHamiltonian)
        self.parent = None  # type: 'QHamiltonian'

    def __len__(self):
        """
        Return the number of all waves from all operators.
        """
        counts = 0
        for key in self._waves.keys():
            counts += len(self._waves[key])
        return counts

    def __getitem__(self, key):
        """
        Return the Item.
        """
        return self._waves[key]

    def __add__(self, other):
        """
        Add two QJob Object, combine the operators with the waves into a new QJob obj.
        """
        newJob = copy.deepcopy(self)
        newJob.appendJob(other)
        return newJob

    def __str__(self) -> str:
        """
        Output the descriptions of the object.
        """
        returnStr = f"QJob ({'No title' if self.title == '' else self.title}):\n"
        for opKey in self._waves.keys():
            returnStr += f"- {opKey}  ({len(self._waves[opKey])} waveforms)\n"
            for wave in self._waves[opKey]:
                waveName = "No name" if wave.name is None or wave.name == '' else wave.name
                seqLen = f"{len(wave.seq)} pieces" if wave.seq is not None else 'None'
                returnStr += f"    - [{wave.tag}] {waveName}: t0={wave.t0}, t={wave.t}, args={wave.args}, " \
                             f"seq={seqLen}, strength={wave.strength}, freq={wave.freq}, " \
                             f"phase={wave.phase}.\n"

        return returnStr

    @property
    def subSysNum(self) -> int:
        """
        Return the number of the subsystems.
        """
        return self._subSysNum

    @subSysNum.setter
    def subSysNum(self, value: int):
        """
        Set the number of the subsystems.

        :param value: the number of the subsystems.
        """
        if not isinstance(value, int):
            raise Error.ArgumentError("subSysNum must be an integer!")
        if value < 1:
            raise Error.ArgumentError("subSysNum must be larger than 0!")
        self._subSysNum = value

    @property
    def sysLevel(self) -> Union[int, List[int]]:
        """
        Return the size of the subsystem.
        """
        return self._sysLevel

    @sysLevel.setter
    def sysLevel(self, value: Union[int, List[int]]):
        """
        Set the level of the subsystems.

        :param value: the number of the subsystems
        """
        if not isinstance(value, int) and not isinstance(value, list):
            raise Error.ArgumentError("sysSize must be an integer or a list!")
        if isinstance(value, list) and min(value) < 2:
            raise Error.ArgumentError("All items in sysSize must be larger than 1!")
        if isinstance(value, int) and value < 2:
            raise Error.ArgumentError(f"System level must be larger than 1, but {value} input!")
        self._sysLevel = value

    @property
    def waveCache(self) -> Dict[str, Any]:
        """
        Return the cache of the pulses.
        """
        return self._waveCache

    @waveCache.deleter
    def waveCache(self):
        del self._waveCache

    @property
    def LO(self) -> Dict[str, Any]:
        """
        Return the cache of the pulses.
        """
        return self._LO

    @LO.setter
    def LO(self, value: Dict[str, Tuple[float, float]]):
        self._LO = copy.deepcopy(value)

    @property
    def waves(self) -> Dict[str, Union[QWaveform, List[QWaveform]]]:
        """
        Return the size of the subsystem.
        """
        return self._waves

    @property
    def ctrlOperators(self) -> Dict[str, Union[QOperator, List[QOperator]]]:
        """
        Return the dictionary of control operators.
        """
        return self._ctrlOperators

    @ctrlOperators.setter
    def ctrlOperators(self, value: Dict[str, Union[QOperator, List[QOperator]]]):
        """
        Set control operators.
        """
        # Check system information
        for opKey in value.keys():
            if isinstance(value[opKey], list):
                for op in value[opKey]:
                    if op.onSubSys + 1 > self._subSysNum:
                        raise Error.ArgumentError(f"Operator `{opKey}` is on the subsystem "
                                                  f"`{value[opKey].onSubSys}`, exceeds the s"
                                                  f"ubSysNum {self._subSysNum}.")

            elif value[opKey].onSubSys + 1 > self._subSysNum:
                raise Error.ArgumentError(f"Operator `{opKey}` is on the subsystem "
                                          f"`{value[opKey].onSubSys}`, exceeds the s"
                                          f"ubSysNum {self._subSysNum}.")

            if isinstance(self._sysLevel, int):
                if isinstance(value[opKey], list):
                    for op in value[opKey]:
                        opLvl = max(op.matrix.shape)
                        if opLvl != self._sysLevel:
                            raise Error.ArgumentError(f"Size of the operator `{opKey}` ({opLvl}) does not"
                                                      f"match with the system level ({self.sysLevel}).")
                else:
                    opLvl = max(value[opKey].matrix.shape)
                    if opLvl != self._sysLevel:
                        raise Error.ArgumentError(f"Size of the operator `{opKey}` ({opLvl}) does not"
                                                  f"match with the system level ({self.sysLevel}).")

            elif isinstance(self._sysLevel, list):
                if isinstance(value[opKey], list):
                    for op in value[opKey]:
                        opLvl = max(op.matrix.shape)
                        if opLvl != self._sysLevel[op.onSubSys]:
                            raise Error.ArgumentError(f"Size of the operator `{opKey}` ({opLvl}) does not"
                                                      f"match with the system level ({self.sysLevel}).")
                else:
                    opLvl = max(value[opKey].matrix.shape)
                    if opLvl != self._sysLevel[value[opKey].onSubSys]:
                        raise Error.ArgumentError(f"Size of the operator `{opKey}` ({opLvl}) on subsystem "
                                                  f"{value[opKey].onSubSys} does not match with the system "
                                                  f"level ({self.sysLevel}).")
        # Copy ctrl operators
        self._ctrlOperators = copy.deepcopy(value)
        # Initialize wave information
        for opKey in self._ctrlOperators.keys():
            if opKey not in self._waves:
                self._waves[opKey] = []

    @ctrlOperators.deleter
    def ctrlOperators(self):
        del self._ctrlOperators
        self._ctrlOperators = {}

    @property
    def dt(self) -> float:
        """
        Return the arbitrary wave generator's (AWG) sampling time interval (also stands for the step size for
        the piecewise-constant quantum simulation algorithm).
        """
        return self._dt

    @dt.setter
    def dt(self, value: float):
        """
        Set the arbitrary wave generator's (AWG) sampling time interval (also stands for the step size for
        the piecewise-constant quantum simulation algorithm).
        """
        if not isinstance(value, float):
            raise Error.ArgumentError("dt must be a float number!")
        if value < 0.0:
            raise Error.ArgumentError("dt must be larger than 0!")
        self._dt = value

    def appendWave(self, operators: Union[QOperator, Callable, List[QOperator], List[Callable]] = None,
                   onSubSys: Union[int, List[int]] = None, waves: Union[QWaveform, List[QWaveform]] = None,
                   strength: Union[int, float] = 1.0, freq: Optional[Union[int, float]] = None,
                   phase: Optional[Union[int, float]] = None, phase0: Optional[Union[int, float]] = None,
                   name: str = None, tag: str = None, shift: float = 0.0, compact: bool = True):
        """
        This method appends control terms and waveforms to a QJob object. Unlike `addWave()`, this function will append
        the waveform in the end of the existing waveforms, hence ignore the `t0` parameter of rhe `QWaveform` object.

        :param operators: wave operator
        :param onSubSys: what subsystem the wave is acting upon
        :param waves: a QWaveform object
        :param strength: wave strength
        :param freq: wave frequency shift
        :param phase: pulse phase shift (will accumulate during the entire pulse execution)
        :param phase0: pulse phase shift (will not accumulate during the entire pulse execution)
        :param name: a user-given name to the wave
        :param tag: wave tag indicating purpose
        :param shift: time shift of the added wave
        :param compact: the added wave will be left aligned in specified control terms if True.
        :return: None
        """
        # When searching for the waves, we do not filter the waves by the tags.
        wavesDict = self.searchWave(operators, onSubSys, names=name)
        maxTime = 0.
        if compact:
            # When compact is True, the wave will be appended to the final time of the existed waves from the
            #     specific control term.
            for waveKey in wavesDict.keys():
                if len(wavesDict[waveKey]) > 0:
                    for waveform in wavesDict[waveKey]:
                        if waveform.flag & QWAVEFORM_FLAG_ALWAYS_ON > 0:
                            continue
                        if waveform.seq is not None:
                            finalNs = waveform.t0 + len(waveform.seq) * self.dt
                        else:
                            finalNs = waveform.t0 + waveform.t
                        if maxTime < finalNs:
                            maxTime = finalNs
        else:
            # When compact is False, the wave will the appended to the final time of all existed waves of all
            #     control terms.
            maxTime, _ = self.computeMaxTime()

        # Finally, the start time of the appended wave is the maxTime plus the time shift value.
        t0 = maxTime + shift
        self.addWave(operators, onSubSys, waves, t0, strength, freq, phase, phase0, name, tag, None)

    def addWave(self, operators: Union[QOperator, Callable, List[QOperator], List[Callable]] = None,
                onSubSys: Union[int, List[int]] = None, waves: Union[QWaveform, List[QWaveform]] = None,
                t0: Union[int, float] = None, strength: Union[int, float] = 1.0,
                freq: Optional[Union[int, float]] = None, phase: Optional[Union[int, float]] = None,
                phase0: Optional[Union[int, float]] = None, name: str = None, tag: str = None, flag: int = None) \
            -> None:
        """
        This method adds control terms and waveforms to a QJob object. Users can specify the operators(QOperator object)
        , subsystem's indexes, wave information (QWaveform object), wave strength, wave frequency, wave phase. Moreover,
        the users can specify a tag and flag for identifying the purposes of the wave.

        :param operators: wave operator
        :param onSubSys: what subsystem the wave is acting upon
        :param waves: a QWaveform object
        :param t0: start time
        :param strength: wave strength
        :param freq: wave frequency shift
        :param phase: pulse phase shift (will accumulate during the entire pulse execution)
        :param phase0: pulse phase shift (will not accumulate during the entire pulse execution)
        :param name: a user-given name to the wave
        :param tag: wave tag indicating purpose
        :param flag: wave flag
        :return: None
        """

        def _addOperator(_flag):
            """ Add Operator """
            if operators is not None and onSubSys is not None:
                # Generate the operator key
                _opKey = self.addOperators(operators, onSubSys, name=name,
                                           flagDoNotClear=(_flag & QWAVEFORM_FLAG_DO_NOT_CLEAR) > 0)
            elif name is not None:
                _opKey = name
            else:
                raise Error.ArgumentError(
                    "You must indicate the operators with onSubSys or the name of the term.")
            return _opKey

        def _getFlagAndWaveContainer(_wave):
            """ Return Flag or wave container. """
            _realFlag = 0
            if flag is not None:
                _realFlag = flag
            elif _wave.flag is not None:
                _realFlag = _wave.flag
            if (_realFlag & QWAVEFORM_FLAG_DO_NOT_CLEAR) and self.parent is None:
                raise Error.ArgumentError("Current QJob object has no parent, hence cannot add "
                                          "waves with QWAVEFORM_FLAG_DO_NOT_CLEAR.")
            elif _realFlag & QWAVEFORM_FLAG_DO_NOT_CLEAR:
                _containerWaves = self.parent.doNotClearFlagWaves
            else:
                _containerWaves = self._waves
            return _realFlag, _containerWaves

        _waves = waves.copy()
        if isinstance(_waves, list):
            for wave in _waves:
                _flag, containerWaves = _getFlagAndWaveContainer(wave)
                opKey = _addOperator(_flag)
                wave.dt = self.dt
                wave.t0 = wave.t0 if t0 is None else t0
                wave.strength = wave.strength if strength is None else strength
                wave.freq = wave.freq if freq is None else freq
                wave.phase = wave.phase if phase is None else phase
                wave.phase0 = wave.phase0 if phase0 is None else phase0
                wave.tag = wave.tag if tag is None else tag
                wave.flag = _flag
                containerWaves[opKey].append(wave)
        elif isinstance(_waves, QWaveform):
            _flag, containerWaves = _getFlagAndWaveContainer(_waves)
            opKey = _addOperator(_flag)
            _waves.dt = self.dt
            _waves.t0 = _waves.t0 if t0 is None else t0
            _waves.strength = _waves.strength if strength is None else strength
            _waves.freq = _waves.freq if freq is None else freq
            _waves.phase = _waves.phase if phase is None else phase
            _waves.phase0 = _waves.phase0 if phase0 is None else phase0
            _waves.tag = _waves.tag if tag is None else tag
            _waves.flag = _flag
            containerWaves[opKey].append(_waves)
        else:
            raise Error.ArgumentError("Parameter waves must be a (or a list of) QWaveform object(s).")

        # Compute the end time
        self.computeMaxTime()
        self.clearCache()

    def addWaveRot(self, onSubSys: int, waves: Union[QWaveform, List[QWaveform]], t0: Union[int, float] = None,
                   detuning: Union[int, float] = 0.0, phase: Union[int, float] = 0.0, tag: str = None) -> None:
        """
        Add the control terms for the two-qubit cross-resonance effect in
        the rotating frame with the frequency of the diagonal elements of drift term.

        :param onSubSys: what subsystem the wave is acting upon
        :param waves: a QWaveform object
        :param t0: start time
        :param detuning: wave's corresponding detuning
        :param phase: wave phase
        :param tag: wave tag
        :return: None
        """
        levelList = []
        if isinstance(self.sysLevel, list):
            levelList = self.sysLevel
        elif isinstance(self.sysLevel, int):
            levelList = [self.sysLevel for _ in range(self.subSysNum)]

        self.addWave(operators=driveX(levelList[onSubSys]), onSubSys=onSubSys, t0=t0,
                     waves=copy.deepcopy(waves), freq=detuning, phase0=phase, tag=tag)
        self.addWave(operators=driveY(levelList[onSubSys]), onSubSys=onSubSys, t0=t0,
                     waves=copy.deepcopy(waves), freq=detuning, phase0=phase + pi / 2, tag=tag)

    def setLO(self, operators: Union[QOperator, Callable, List[QOperator], List[Callable]],
              onSubSys: Union[int, List[int]], name: str = None, freq: float = 1.0, phase: float = 0.0):
        """
        Add local oscillator for specific control terms.

        :param operators: QOperator object(s)
        :param onSubSys: what subsystem the wave is acting upon
        :param name: user-defined operator name
        :param freq: the frequency of the local oscillator
        :param phase: the phase between the local oscillator and the waveform
        """
        # Add the operator
        opKey = self.addOperators(operators, onSubSys, name)
        # Set the local oscillator frequency and the phase
        self._LO[opKey] = (freq, phase)

    def addOperators(self, operators: Union[QOperator, Callable, List[QOperator], List[Callable]],
                     onSubSys: Union[int, List[int]], name: str = None, flagDoNotClear: bool = False) -> str:
        """
        Add control terms onto the specified subsystem.

        :param operators: QOperator object(s)
        :param onSubSys: what subsystem the wave is acting upon
        :param name: user-defined operator name
        :param flagDoNotClear: whether this operator will be cleared
        """

        _operators = formatOperatorInput(operators, onSubSys, self.sysLevel)

        if _operators is not None and onSubSys is None:
            # Input the complete matrices directly.
            if isinstance(self._sysLevel, int):
                dim = self._sysLevel ** self._subSysNum
            elif isinstance(self._sysLevel, list):
                dim = 1
                for lvl in self._sysLevel:
                    dim = dim * lvl
            else:
                raise Error.ArgumentError(f"Unsupported type of self._sysLevel ({type(self._sysLevel)}).")

            if not shape(_operators.matrix) == (dim, dim):
                raise Error.ArgumentError(f"Dimension of operator ({dim}, {dim}) does not match the sysLevel.")

            operatorForSave = _operators
        else:
            # If users input the onSubSys, combineOperatorAndOnSubSys will update the onSubSys of
            #     the operators.
            operatorForSave = combineOperatorAndOnSubSys(self._subSysNum, _operators, onSubSys)

        # Set the name if it is given
        if name is not None:
            opKey = name
        else:
            # Use function generateOperatorKey to generate the key of the operators.
            opKey = generateOperatorKey(self.subSysNum, operatorForSave)

        if flagDoNotClear and self.parent is None:
            raise Error.ArgumentError("Current QJob object has no parent, hence cannot add "
                                      "waves with QWAVEFORM_FLAG_DO_NOT_CLEAR.")
        elif not flagDoNotClear:
            containerCtrlOperators = self._ctrlOperators
            containerWaves = self._waves
        else:
            containerCtrlOperators = self.parent.doNotClearFlagOperators
            containerWaves = self.parent.doNotClearFlagWaves

        # Verify the uniqueness of the combination of the name and onSubSys
        if opKey not in containerCtrlOperators.keys():
            # Save the operators.
            containerCtrlOperators[opKey] = operatorForSave
            containerWaves[opKey] = []

        # Compute the end time
        self.computeMaxTime()

        return opKey

    def appendJob(self, obj: Any, t0: float = None) -> None:
        """
        Append objects to the QJob object.
        When timeShift is None, all waves appended will be started from the end of the current waves.

        :param obj: object to be added
        :param t0: start time
        :return: None
        """
        if isinstance(obj, QJob):
            # When the object is a QJob.
            if obj.sysLevel != self.sysLevel:
                raise Error.ArgumentError("sysLevel does not match!")
            if obj.subSysNum != self.subSysNum:
                raise Error.ArgumentError("subSysNum does not match!")
            # Start adding
            self.computeMaxTime()
            if t0 is None:
                _timeShift = self.endTime
            else:
                _timeShift = t0
            _obj = copy.deepcopy(obj)
            for key in _obj.ctrlOperators.keys():
                if key not in self._ctrlOperators.keys():
                    # Current term exists, then add the wave directly.
                    self._ctrlOperators[key] = copy.deepcopy(_obj.ctrlOperators[key])
                    self._waves[key] = []
                # Then add the wave.
                for wave in _obj.waves[key]:
                    newWave = copy.deepcopy(wave)
                    newWave.t0 += _timeShift
                    self._waves[key].append(newWave)
            self.computeMaxTime()
        else:
            raise Error.ArgumentError(f"Unsupported type ({type(obj)}), only QJob is supported.")

    def clearWaves(self, operators: Union[QOperator, Callable, List[QOperator], List[Callable]] = None,
                   onSubSys: Union[int, List[int]] = None, names: Union[str, List[str]] = None,
                   tag: str = None) -> None:
        """
        Remove all waveforms in the specified control terms.
        If no names are given, remove all waveforms in all control terms.

        :param operators: the operator of the wave
        :param onSubSys: the subsystem of the wave
        :param names: the name of control term
        :param tag: the tag of the waves
        :return: None
        """
        # Generate opKeys of the waveforms which are going to be cleared
        if names is not None:
            # Set the name if it is given
            opKey = names
        elif operators is not None and onSubSys is not None:
            _operators = formatOperatorInput(operators, onSubSys, self.sysLevel)
            # If users input the onSubSys, combineOperatorAndOnSubSys will update the onSubSys of
            #     the operators.
            operatorForSave = combineOperatorAndOnSubSys(self._subSysNum, _operators, onSubSys)
            # Use function generateOperatorKey to generate the key of the operators.
            if isinstance(_operators, list):
                opKey = []
                for op in _operators:
                    opKey.append(generateOperatorKey(self.subSysNum, op))
            else:
                opKey = generateOperatorKey(self.subSysNum, operatorForSave)
        else:
            # Remove all waveforms
            opKey = self._ctrlOperators.keys()

        # Generate opKeys of the waveforms which are going to be cleared
        tagBlackList = []
        if tag is not None:
            if isinstance(tag, list):
                tagBlackList.extend(tag)
            else:
                tagBlackList.append(tag)

        # Generate new wave list
        newWaves = {}
        newCtrlOperators = {}
        for key in self._waves.keys():
            _toSave = []
            for wave in self._waves[key]:
                _ifSave = False
                # Check flag QWAVEFORM_DO_NOT_CLEAR
                _ifSave = _ifSave or ((wave.flag & QWAVEFORM_FLAG_DO_NOT_CLEAR) > 0)
                # Check opKey
                _ifSave = _ifSave or (key not in opKey)
                # Check tag
                _ifSave = _ifSave or (len(tagBlackList) > 0 and wave.tag not in tagBlackList)
                if _ifSave:
                    _toSave.append(copy.deepcopy(wave))
            if len(_toSave) > 0:
                newWaves[key] = _toSave
                newCtrlOperators[key] = copy.deepcopy(self._ctrlOperators[key])

        del self._waves
        del self._ctrlOperators
        self._waves = newWaves
        self._ctrlOperators = newCtrlOperators
        self.computeMaxTime()

    def searchWave(self, operators: Union[QOperator, Callable, List[QOperator], List[Callable]] = None,
                   onSubSys: Union[int, List[int]] = None, names: Union[str, List[str]] = None, tag: str = None):
        """
        Search waves by different methods.

        :param operators: search by operator
        :param onSubSys: search by subsystem
        :param names: search by name
        :param tag: search by tag
        """
        return self.searchWaveTool(self._waves, self._subSysNum, self._sysLevel, operators, onSubSys, names, tag)

    @staticmethod
    def searchWaveTool(container: Dict[str, List[QWaveform]], subSysNum: int, sysLevel: Union[int, List[int]],
                       operators: Union[QOperator, Callable, List[QOperator], List[Callable]] = None,
                       onSubSys: Union[int, List[int]] = None, names: Union[str, List[str]] = None,
                       tag: str = None) -> Dict[str, List[QWaveform]]:
        """
        Search waves by different methods.

        :param container: input dictionary
        :param subSysNum: subsystem number
        :param sysLevel: the system levels
        :param operators: search by operator
        :param onSubSys: search by subsystem
        :param names: search by name
        :param tag: search by tag
        """

        returnDict = {}
        if names is not None:
            def _searchWaveByNames(_returnDict, _name):
                """ Search Waves by name and tag """
                if _name in container.keys():
                    if tag is None:
                        _returnDict[_name] = container[_name]
                    else:
                        _returnDict[_name] = []
                        for _wave in container[_name]:
                            if _wave.tag == tag:
                                _returnDict[_name].append(_wave)

            # When names (of the operators) is not None, then find the waves by the name string
            if isinstance(names, str):
                # If user input a name
                _searchWaveByNames(returnDict, names)
                return returnDict
            elif isinstance(names, list):
                # If user input a list of names
                for name in names:
                    _searchWaveByNames(returnDict, name)
                return returnDict
            else:
                raise Error.ArgumentError("Variable names should be a list or int.")
        elif operators is not None and onSubSys is not None:
            # When names (of the operators) is None, then find the waves by the operators with onSubSys.
            formattedOperators = formatOperatorInput(operators, onSubSys, sysLevel)
            if isinstance(operators, list) and isinstance(onSubSys, list):
                _operators = formattedOperators
                _onSubSys = onSubSys
            else:
                _operators = [formattedOperators]
                _onSubSys = [onSubSys]

            if len(_operators) != len(_onSubSys):
                raise Error.ArgumentError("Length of operators and onSubSys must be the same.")

            for idx in range(len(_operators)):
                operatorForSearch = combineOperatorAndOnSubSys(subSysNum, _operators, _onSubSys)
                opKey = generateOperatorKey(subSysNum, operatorForSearch)
                if opKey in container.keys():
                    if tag is None:
                        returnDict[opKey] = container[opKey]
                    else:
                        returnDict[opKey] = []
                        for wave in container[opKey]:
                            if tag == wave.tag:
                                returnDict[opKey].append(wave)
            return returnDict
        elif tag is not None:
            for opKey in container.keys():
                returnDict[opKey] = []
                for waveform in container[opKey]:
                    if waveform.tag == tag:
                        returnDict[opKey].append(waveform)
            return returnDict
        else:
            raise Error.ArgumentError("You must indicate the operators with onSubSys / names / tag.")

    def plot(self, names: Union[str, List[str]] = None, tag: str = None, xUnit: str = 'ns', yUnit: str = 'Amp (a.u.)',
             color: Union[str, List[str]] = None, dark: bool = False) -> None:
        """
        Print the waveforms of the control terms listed in ``names``.

        :param names: the name or name list of the control term
        :param tag: the tag of the waveforms to be plotted.
        :param xUnit: the unit of the x axis
        :param yUnit: the unit of the y axis
        :param color: None or list of colors
        :param dark: the plot can be switched to dark mode if required by user
        :return: None

        In Quanlse, users can specify colors from:

        ``mint``, ``blue``, ``red``, ``green``, ``yellow``, ``black``, ``pink``, ``cyan``, ``purple``,
        ``darkred``, ``orange``, ``brown``, ``pink`` and ``teal``.

        The colors will repeat if there are more pulses than colors.
        """

        self.computeMaxTime()

        # Get the max time
        maxNs = self.endTime

        # print plot
        if names is None:
            nameList = []
            nameList.extend(self.waves.keys())
            if self.parent is not None:
                nameList.extend(self.parent.doNotClearFlagWaves.keys())
            names = sorted(nameList)
        elif isinstance(names, str):
            names = [names]
        elif isinstance(names, list):
            pass
        else:
            raise Error.ArgumentError("Variable names should be a list or str.")

        if len(names) < 1:
            raise Error.ArgumentError("No terms exist, thus no waves can be plotted.")

        # Keep track of figure numbers
        fig = 0

        # Create two empty lists
        x = []
        y = []
        yLabel = []
        colors = []
        colorIndex = 0
        counting = self.buildWaveCache(tag=tag, forPlot=True)
        # Plot waves
        plotNames = []
        isComplex = False
        for name in names:
            if name not in counting.keys() or counting[name] < 1:
                continue
            aList = array(self._waveCache[name])
            tList = linspace(0, maxNs, len(aList))
            y.append(list(aList))
            x.append(list(tList))
            yLabel.append(yUnit)
            plotNames.append(name)
            # If the amplitues are complex number
            if aList.dtype in [numpy.complex, numpy.complex128]:
                isComplex = True
            # Whether repetitive colors or all blue
            if color is None:
                colors.append('blue')
            else:
                colors.append(color[colorIndex])
                colorIndex += 1
                if colorIndex == len(color):
                    colorIndex = 0
            fig += 1
        plotPulse(x, y, xLabel=f'Time ({xUnit})', yLabel=yLabel, title=plotNames,
                  color=colors, dark=dark, complx=isComplex)
        plt.show()
        self.clearCache()

    def plotIon(self, names: Union[str, List[str]] = None, xUnit: str = r'$\mu$s', yUnit: str = 'Amp (a.u.)',
                color: Union[str, List[str]] = None, dark: bool = False) -> None:
        """
        Print the waveforms of the control terms listed in ``names``.

        :param names: the name or name list of the control term
        :param xUnit: the unit of the x axis
        :param yUnit: the unit of the y axis
        :param color: None or list of colors
        :param dark: the plot can be switched to dark mode if required by user
        :return: None

        In Quanlse, users can specify colors from:

        ``mint``, ``blue``, ``red``, ``green``, ``yellow``, ``black``, ``pink``, ``cyan``, ``purple``,
        ``darkred``, ``orange``, ``brown``, ``pink`` and ``teal``.

        The colors will repeat if there are more pulses than colors.
        """

        self.computeMaxTime()

        # Get the max time
        maxNs = self.endTime

        # print plot
        if names is None:
            names = list(self.waves.keys())
        elif isinstance(names, str):
            names = [names]
        elif isinstance(names, list):
            pass
        else:
            raise Error.ArgumentError("Variable names should be a list or a str.")

        if len(names) < 1:
            raise Error.ArgumentError("No terms exist, hence no available waves can be plot.")

        # Keep track of figure numbers
        fig = 0

        # Create two empty lists
        x = []
        y = []
        yLabel = []
        colors = []
        colorIndex = 0
        self.buildWaveCache()

        for name in names:
            aList = array(self._waveCache[name])
            tList = linspace(0, maxNs, len(aList))
            y.append(list(aList))
            x.append(list(tList))
            yLabel.append(yUnit)

            # Whether repetitive colors or all blue
            if color is None:
                colors.append('blue')
            else:
                colors.append(color[colorIndex])
                colorIndex += 1
                if colorIndex == len(color):
                    colorIndex = 0
            fig += 1
        plotPulse(x, y, xLabel=f'Time ({xUnit})', yLabel=yLabel, title=names, color=colors, dark=dark)
        plt.show()
        self.clearCache()

    def keys(self) -> KeysView:
        """
        Return all the keys of waves.

        :return: returned keys
        """
        return self._waves.keys()

    def computeMaxTime(self) -> Tuple[float, float]:
        """
        Compute the time duration of the whole circuit according to the waves added.

        :return: returned total time duration
        """
        # Find the longest time
        maxTime = 0.0
        containers = [self._waves]
        if self.parent is not None:
            containers.append(self.parent.doNotClearFlagWaves)
        for container in containers:
            for key in container.keys():
                waves = container[key]
                for waveform in waves:
                    if waveform.flag & QWAVEFORM_FLAG_ALWAYS_ON > 0:
                        continue
                    if waveform.seq is not None:
                        finalNs = waveform.t0 + len(waveform.seq) * self.dt
                    else:
                        finalNs = waveform.t0 + waveform.t
                    if maxTime < finalNs:
                        maxTime = finalNs
        maxDt = floor(maxTime / self.dt)

        self.endTimeDt = maxDt
        self.endTime = maxTime
        return maxTime, maxDt

    def clearCache(self) -> None:
        """
        Clear the wave cache.

        :return: None
        """
        self._waveCache = {}

    def buildWaveCache(self, tag: str = None, forPlot: bool = False) -> Dict[str, int]:
        """
        Generate the pulse sequences for further usage.

        :param tag: wave tag
        :param forPlot: whether plot
        :return: returned pulse sequences
        """
        # Generate the pulse sequences for all of the control terms.
        waveCounting = {}
        self.clearCache()
        allKeys = list(self.waves.keys())
        if self.parent is not None:
            allKeys.extend(self.parent.doNotClearFlagWaves.keys())
        for key in allKeys:
            _sequence = self.generatePulseSequence(name=key, tag=tag, forPlot=forPlot, counting=waveCounting)
            self._waveCache[key] = _sequence
        return waveCounting

    def generatePulseSequence(self, operators: Union[QOperator, Callable, List[QOperator], List[Callable]] = None,
                              onSubSys: Union[int, List[int]] = None, name: str = None, tag: str = None,
                              forPlot: bool = False, counting: Dict[str, int] = None,
                              phaseList: List[int] = None) -> List[Union[float, complex]]:
        """
        Generate the piecewise-constant pulse sequence according to the waveform configurations during
        the whole evolution time.

        :param operators: the list of operators
        :param onSubSys: the list of subsystems
        :param name: the key of control term
        :param tag: the tag of the waveforms
        :param forPlot: whether the sequence is for plotting
        :param counting: the waveform counting for all control terms
        :param phaseList: the list of phase on every time piece

        :return: the list of pulse sequence
        """

        _sequenceList = []
        _phaseList = []
        self.computeMaxTime()

        # Generate waveKey
        if name is not None:
            waveKey = name
        elif operators is not None and onSubSys is not None:
            _operators = formatOperatorInput(operators, onSubSys, self._sysLevel)
            operatorForSearch = combineOperatorAndOnSubSys(self._subSysNum, _operators, onSubSys)
            waveKey = generateOperatorKey(self.subSysNum, operatorForSearch)
        else:
            raise Error.ArgumentError("You must indicate the operators with onSubSys or the name of the term.")

        # Combine all waves (waves from current QJob instance and the parent QHamiltonian instance)
        combinedWaves = []
        if waveKey in self.waves:
            combinedWaves.extend(copy.deepcopy(self.waves[waveKey]))
        if self.parent is not None and waveKey in self.parent.doNotClearFlagWaves.keys():
            combinedWaves.extend(copy.deepcopy(self.parent.doNotClearFlagWaves[waveKey]))

        # A register for recording the phase accumulation for every control term.
        #     The values in this register may change over time.
        phaseAcc = {}
        phaseAccList = []

        # Traverse all the time slices.
        for nowDt in range(self.endTimeDt):
            # The amount of phase
            finalPhase = 0.
            # The amount of amplitude
            currentAmp = 0.
            # The nanosecond time of this moment
            nowNs = nowDt * self.dt + self.dt / 2.
            # The phase waveforms can work in this time window
            timeWindow = (nowNs - self.dt / 2., nowNs + self.dt / 2.)
            # Traverse all the waveforms
            for waveId, waveform in enumerate(combinedWaves):
                if waveform.flag & QWAVEFORM_FLAG_DO_NOT_PLOT > 0 and forPlot is True:
                    continue
                elif waveform.flag & QWAVEFORM_FLAG_ALWAYS_ON > 0:
                    waveform.t0 = 0.0
                    waveform.t = self.endTime
                else:
                    if tag is not None and waveform.tag != tag:
                        continue

                # Count the number of waveforms of the given operator
                if counting is not None:
                    if waveKey not in counting.keys():
                        counting[waveKey] = 0
                    counting[waveKey] += 1

                # Calculate the start and end time
                insertNs = waveform.t0
                if waveform.seq is not None:
                    endNs = waveform.t0 + len(waveform.seq) * self.dt
                else:
                    endNs = waveform.t0 + waveform.t

                # if the current waveform works
                if waveform.flag & QWAVEFORM_FLAG_PHASE_SHIFT > 0:
                    isWaveWorking = timeWindow[0] <= waveform.t0 < timeWindow[1]
                else:
                    isWaveWorking = insertNs <= nowNs < endNs

                if isWaveWorking:
                    # Calculate the waveforms' amplitudes.
                    waveform.dt = self.dt
                    if waveKey in self._LO.keys() or any([waveform.freq, waveform.phase0, waveform.phase]):
                        # The freq and phase shift set in QWaveform object
                        freq = 0. if waveform.freq is None else waveform.freq
                        phase = 0. if waveform.phase is None else waveform.phase
                        phase0 = 0. if waveform.phase0 is None else waveform.phase0
                        # Calculate the phase of the contemporary waveform
                        totalPhase = phase0
                        totalFreq = freq
                        # Plus the phase and frequency of the LO setup in QJob instance.
                        if waveKey in self._LO.keys():
                            totalFreq += self._LO[waveKey][0]
                            totalPhase += self._LO[waveKey][1]
                        # Calculate the accumulated phase in phaseAcc
                        if waveKey not in phaseAcc.keys():
                            phaseAcc[waveKey] = phase
                        else:
                            if waveId not in phaseAccList:
                                phaseAcc[waveKey] += phase
                                phaseAccList.append(waveId)
                        # The constant freq and phase set in QJob
                        totalPhase += phaseAcc[waveKey]
                        # Calculate the final LO value
                        if waveform.flag & QWAVEFORM_FLAG_MIX_WAVE > 0:
                            # If the waveform is the mixture of X and Y control
                            amp, varPhase = waveform(nowNs - waveform.t0)
                            finalPhase = totalPhase + varPhase
                            lo = math.cos(totalFreq * nowNs + finalPhase)
                            currentAmp += amp * lo
                        else:
                            finalPhase = totalPhase
                            lo = math.cos(totalFreq * nowNs + finalPhase)
                            currentAmp += waveform(nowNs - waveform.t0) * lo
                    else:
                        currentAmp += waveform(nowNs - waveform.t0)

                    # Modify the amplitudes by drive strength and amplitude noise.
                    currentAmp = currentAmp * waveform.strength
            _phaseList.append(finalPhase)
            _sequenceList.append(currentAmp)

        if phaseList is not None:
            phaseList = _phaseList
        del combinedWaves
        return _sequenceList

    def waveFunctionsToSequences(self, maxEndTime: float = None) -> 'QJob':
        """
        Convert all functions included QWaveform objects into a serializable sequence.

        :param maxEndTime: maximum ending time
        :return: returned QJob object
        """
        obj = copy.deepcopy(self)
        obj.clearWaves()
        for waveKey in self.waves.keys():
            # Add QOperator
            qOp = self.ctrlOperators[waveKey]
            if isinstance(qOp, list):
                onSubSys = []
                for op in qOp:
                    onSubSys.append(op.onSubSys)
            else:
                onSubSys = qOp.onSubSys
            obj.addOperators(qOp, onSubSys, waveKey)
            # Add QWaveform
            for wave in self.waves[waveKey]:
                seqWave = wave.waveFunctionToSequence(obj.dt, maxEndTime=maxEndTime)
                obj.addWave(name=waveKey, waves=seqWave, tag=wave.tag, strength=wave.strength, freq=wave.freq,
                            phase=wave.phase)
        return obj

    def dump(self, maxEndTime: float = None) -> str:
        """
        Return base64 encoded string.

        :param maxEndTime: maximum ending time
        :return: a base64 encoded string
        """

        # The critical step is transforming all waveform functions to the
        #     sequence, because functions can not the serialization.
        obj = self.waveFunctionsToSequences(maxEndTime=maxEndTime)
        obj.parent = None
        # For compatibility with previous versions
        for opKey in obj.waves.keys():
            for wave in obj.waves[opKey]:
                if not hasattr(wave, "omega"):
                    wave.omega = wave.freq
                    wave.phi = 0.
                if not hasattr(wave, "phi"):
                    wave.phi = 0. if wave.phase0 is None else wave.phase0
        # Dump the object
        byteStr = pickle.dumps(obj)
        base64str = base64.b64encode(byteStr)
        del obj
        return base64str.decode()

    def copy(self) -> 'QJob':
        """
        Return the copy of the object
        """
        return copy.deepcopy(self)

    @staticmethod
    def load(base64Str: str) -> 'QJob':
        """
        Create object from base64 encoded string.

        :param base64Str: a base64 encoded string
        :return: a QJob object
        """
        byteStr = base64.b64decode(base64Str.encode())
        obj = pickle.loads(byteStr)  # type: QJob

        # For compatibility with previous versions
        for opKey in obj.waves.keys():
            for wave in obj.waves[opKey]:
                if not hasattr(wave, "freq"):
                    wave.freq = wave.omega if hasattr(wave, "omega") else None
                if not hasattr(wave, "phase0"):
                    wave.phase0 = wave.phi if hasattr(wave, "phi") else None
                if not hasattr(wave, "phase"):
                    wave.phase = None

        return obj


class QJobList:
    """
    Class of QJobList

    :param subSysNum: subsystem number
    :param sysLevel: system energy level
    :param dt: AWG sampling time
    :param title: a user-given title
    :param description: a user-given description, preferably indicating the purpose
    """

    def __init__(self, subSysNum: int, sysLevel: Union[int, List[int]], dt: float, title: str = "",
                 description: str = "") -> None:
        """
        Constructor for QJobList class.
        """
        self._jobList = []  # type: List[QJob]
        self.title = title  # type: str
        self.description = description  # type: str
        self.subSysNum = subSysNum  # type: int
        self.sysLevel = sysLevel  # type: Union[int, List[int]]
        self._LO = {}  # type: Dict[str, Tuple[float, float]]
        self.dt = dt

    @property
    def subSysNum(self) -> int:
        """
        Return the number of the subsystems.
        """
        return self._subSysNum

    @subSysNum.setter
    def subSysNum(self, value: int):
        """
        Set the number of the subsystems.

        :param value: the number of the subsystems.
        """
        if not isinstance(value, int):
            raise Error.ArgumentError("subSysNum must be an integer!")
        if value < 1:
            raise Error.ArgumentError("subSysNum must be larger than 0!")
        self._subSysNum = value

    @property
    def sysLevel(self) -> Union[int, List[int]]:
        """
        Return the size of the subsystem.
        """
        return self._sysLevel

    @sysLevel.setter
    def sysLevel(self, value: Union[int, List[int]]):
        """
        Set the level of the subsystems.

        :param value: the number of the subsystems.
        """
        if not isinstance(value, int) and not isinstance(value, list):
            raise Error.ArgumentError("sysSize must be an integer or a list!")
        if isinstance(value, list) and min(value) < 2:
            raise Error.ArgumentError("All items in sysSize must be larger than 1!")
        if isinstance(value, int) and value < 2:
            raise Error.ArgumentError("All items in sysSize must be larger than 1!")
        self._sysLevel = value

    @property
    def LO(self) -> Dict[str, Any]:
        """
        Return the cache of the pulses.
        """
        return self._LO

    @LO.setter
    def LO(self, value: Dict[str, Tuple[float, float]]):
        self._LO = copy.deepcopy(value)

    @property
    def jobs(self) -> List[QJob]:
        return self._jobList

    def createJob(self) -> QJob:
        """
        Return a instance of QJob which has same system properties

        :return: returned QJob object
        """
        newJob = QJob(subSysNum=self.subSysNum, sysLevel=self.sysLevel, dt=self.dt)
        newJob.LO = self.LO
        return newJob

    def addJob(self, jobs: Union[QJob, List[QJob]]) -> None:
        """
        Add QJob(s) object into the job list.

        :param: jobs to be added
        :return: None
        """
        if isinstance(jobs, list):
            for job in jobs:
                if job.sysLevel != self.sysLevel:
                    raise Error.ArgumentError("sysLevel does not match!")
                if job.subSysNum != self.subSysNum:
                    raise Error.ArgumentError("subSysNum does not match!")
                self._jobList.append(copy.deepcopy(job))
        elif isinstance(jobs, QJob):
            if jobs.sysLevel != self.sysLevel:
                raise Error.ArgumentError("sysLevel does not match!")
            if jobs.subSysNum != self.subSysNum:
                raise Error.ArgumentError("subSysNum does not match!")
            self._jobList.append(copy.deepcopy(jobs))
        else:
            raise Error.ArgumentError("Unsupported input.")

    def clearJob(self) -> None:
        """
        Clear all QJob objects.

        :return: None
        """
        self._jobList = []

    def dump(self, maxEndTime: float = None) -> str:
        """
        Return base64 encoded string.

        :param maxEndTime: maximum ending time
        :return: a base64 encoded string
        """

        # The critical step is transforming all waveform functions to the
        #     sequence, because functions can not the serialization.
        obj = copy.deepcopy(self)
        obj.clearJob()
        for job in self._jobList:
            job.parent = None
            obj.addJob(job.waveFunctionsToSequences(maxEndTime=maxEndTime))

        # For compatibility with previous versions
        for job in obj.jobs:
            for opKey in job.waves.keys():
                for wave in job.waves[opKey]:
                    if not hasattr(wave, "omega"):
                        wave.omega = wave.freq
                        wave.phi = 0.
                    if not hasattr(wave, "phi"):
                        wave.phi = 0. if wave.phase0 is None else wave.phase0
        # Dump the object
        byteStr = pickle.dumps(obj)
        base64str = base64.b64encode(byteStr)
        del obj
        return base64str.decode()

    def clone(self) -> 'QJobList':
        """
        Return the copy of the object
        """
        return copy.deepcopy(self)

    @staticmethod
    def formatInputOperators() -> Union[List[QOperator], QOperator]:
        """
        The input operators may be the callables or the QOperator instances,
        """

    @staticmethod
    def load(base64Str: str) -> 'QJobList':
        """
        Create object from base64 encoded string.

        :param base64Str: a base64 encoded string
        :return: a QJobList object
        """
        byteStr = base64.b64decode(base64Str.encode())
        obj = pickle.loads(byteStr)  # type: QJobList
        # For compatibility with previous versions
        for job in obj.jobs:
            for opKey in job.waves.keys():
                for wave in job.waves[opKey]:
                    if not hasattr(wave, "freq"):
                        wave.freq = wave.omega if hasattr(wave, "omega") else None
                    if not hasattr(wave, "phase0"):
                        wave.phase0 = wave.phi if hasattr(wave, "phi") else None
                    if not hasattr(wave, "phase"):
                        wave.phase = None
        return obj


class QResult:
    """
    Class of QResult

    :param title: a user-given title
    :param description: a user-given description, preferably indicating the purpose
    """

    def __init__(self, title: str = "", description: str = "") -> None:
        """
        Constructor for class QResult
        """
        self.title = title
        self.description = description
        self._results = []

        # For iterator
        self._iter_count = 0

    def __len__(self) -> int:
        return len(self._results)

    def __getitem__(self, item: int) -> Any:
        return self._results[item]

    def __setitem__(self, key: int, value: Any):
        self._results[key] = value

    @property
    def result(self) -> Union[List, Dict]:
        return self._results

    def append(self, result: Any) -> None:
        """
        Set result.

        :param result: result to be set to
        :return: None
        """
        if isinstance(result, QResult):
            for item in result:
                self._results.append(item)
        else:
            self._results.append(result)

    def dump(self) -> str:
        """
        Return base64 encoded string.

        :return: a base64 encoded string
        """
        # Dump the object
        byteStr = pickle.dumps(self)
        base64str = base64.b64encode(byteStr)
        return base64str.decode()

    def clone(self) -> 'QResult':
        """
        Return the copy of the object
        """
        return copy.deepcopy(self)

    @staticmethod
    def load(base64Str: str) -> 'QResult':
        """
        Create object from base64 encoded string.

        :param base64Str: a base64 encoded string
        """
        byteStr = base64.b64decode(base64Str.encode())
        obj = pickle.loads(byteStr)  # type: QResult
        return obj


def gaussian(t: Union[int, float], a: float, tau: float, sigma: float, t0: Union[int, float] = 0.,
             freq: float = None, phase: float = None, phase0: float = None) -> QWaveform:
    """
    Return a QWaveform object of Gaussian wave.

    :param t: pulse duration
    :param a: pulse amplitude
    :param tau: pulse center position
    :param sigma: pulse standard deviation
    :param t0: start time
    :param freq: pulse frequency shift (will not accumulate during the entire pulse execution)
    :param phase: pulse phase shift (will accumulate during the entire pulse execution)
    :param phase0: pulse phase shift (will not accumulate during the entire pulse execution)
    :return: returned Gaussian QWaveform
    """

    def func(_t, args):
        """ The Gaussian function """
        _a, _tau, _sigma = args
        if _sigma == 0:
            return 0
        pulse = _a * exp(- ((_t - _tau) ** 2 / (2 * _sigma ** 2)))
        return pulse

    wave = QWaveform(f=func, t0=t0, t=t, args=(a, tau, sigma), freq=freq, phase=phase, phase0=phase0)
    wave.name = "gaussian"
    return wave


def square(t: Union[int, float], a: float, t0: Union[int, float] = 0., freq: float = None,
           phase: float = None, phase0: float = None) -> QWaveform:
    """
    Return a QWaveform object of square wave.

    :param t: pulse duration
    :param a: pulse amplitude
    :param t0: start time
    :param freq: pulse frequency shift (will not accumulate during the entire pulse execution)
    :param phase: pulse phase shift (will accumulate during the entire pulse execution)
    :param phase0: pulse phase shift (will not accumulate during the entire pulse execution)
    :return: returned square QWaveform
    """

    def func(_t, args):
        """ The constant function """
        return args

    wave = QWaveform(f=func, t0=t0, t=t, args=a, freq=freq, phase=phase, phase0=phase0)
    wave.name = "square"
    return wave


def delay(t: Union[int, float], t0: Union[int, float] = 0., freq: float = None,
          phase: float = None, phase0: float = None) -> QWaveform:
    """
    Return a QWaveform object of delay.

    :param t: pulse duration
    :param t0: start time
    :param freq: pulse frequency shift (will not accumulate during the entire pulse execution)
    :param phase: pulse phase shift (will accumulate during the entire pulse execution)
    :param phase0: pulse phase shift (will not accumulate during the entire pulse execution)
    :return: returned square QWaveform
    """

    def func(_t, _):
        """ The constant function """
        return 0.

    wave = QWaveform(f=func, t0=t0, t=t, args=None, freq=freq, phase=phase, phase0=phase0)
    wave.name = "delay"
    return wave


def virtualZ(phase: float = None, t0: Union[int, float] = 0.) -> QWaveform:
    """
    Return a QWaveform object of phase shift (implement Virtual-Z gate).

    :param phase: pulse phase shift (will accumulate during the entire pulse execution)
    :param t0: start time
    :return: returned square QWaveform
    """

    def func(_t, _):
        """ The constant function """
        return 0.

    wave = QWaveform(f=func, t0=t0, t=0.0, args=None, freq=0.0, phase=phase, phase0=0.0,
                     flag=QWAVEFORM_FLAG_PHASE_SHIFT)
    wave.name = "virtualZ"
    return wave


def sin(t: Union[int, float], a: float, b: float, c: float, t0: Union[int, float] = 0.,
        freq: float = None, phase: float = None, phase0: float = None) -> QWaveform:
    """
    Return a QWaveform object of sin wave.
    x(t) = a * sin(b * t + c)

    :param t: pulse duration
    :param a: sin wave amplitude
    :param b: 2 * pi / (sin wave period)
    :param c: sin wave phase
    :param t0: start time
    :param freq: pulse frequency shift (will not accumulate during the entire pulse execution)
    :param phase: pulse phase shift (will accumulate during the entire pulse execution)
    :param phase0: pulse phase shift (will not accumulate during the entire pulse execution)
    :return: returned sin QWaveform
    """

    def func(_t, args):
        """ The sin function """
        _a, _b, _c = args
        return _a * math.sin(_b * _t + _c)

    wave = QWaveform(f=func, t0=t0, t=t, args=(a, b, c), freq=freq, phase=phase, phase0=phase0)
    wave.name = "sin"
    return wave


def sequence(seq: List[Union[float, complex]], t0: Union[int, float] = 0., freq: float = None, phase: float = None,
             phase0: float = None) -> QWaveform:
    """
    Return a QWaveform object of pulse sequence.

    :param seq: pulse sequence
    :param t0: pulse start time
    :param freq: pulse frequency shift (will not accumulate during the entire pulse execution)
    :param phase: pulse phase shift (will accumulate during the entire pulse execution)
    :param phase0: pulse phase shift (will not accumulate during the entire pulse execution)
    :return: returned QWaveform object
    """
    wave = QWaveform(t0=t0, seq=seq, freq=freq, phase=phase, phase0=phase0)
    wave.name = "manual_sequence"
    return wave


def quasiSquareErf(t: Union[int, float], a: float, l: float, r: float, sk: float = None, t0: Union[int, float] = 0.,
                   freq: float = None, phase: float = None, phase0: float = None) -> QWaveform:
    """
    Return the sample pulse with a quasi-square envelope.

    :param t: pulse duration
    :param a: pulse amplitude
    :param l: quasiSquare wave function parameter
    :param r: quasiSquare wave function parameter
    :param sk: quasiSquare wave function parameter
    :param t0: pulse start time
    :param freq: pulse frequency shift shift (will not accumulate during the entire pulse execution)
    :param phase: pulse phase shift (will accumulate during the entire pulse execution)
    :param phase0: pulse phase shift (will not accumulate during the entire pulse execution)
    :return: returned quasi square QWaveform object
    """

    def func(_t, args):
        """ The quasi square error function """
        _a, _l, _r, _sk = args
        if _sk is None:
            _sk = _a * 0.3
        if _a == 0.0:
            return 0.0
        t1i = _l
        t2i = _r
        pulse = 0.25 * _a * (1 + erf(sqrt(pi) * _sk / _a * (_t - t1i)))
        pulse = pulse * (1 - erf(sqrt(pi) * _sk / _a * (_t - t2i)))
        return pulse

    wave = QWaveform(f=func, t0=t0, t=t, args=(a, l, r, sk), freq=freq, phase=phase, phase0=phase0)
    wave.name = "quasi_square_erf"
    return wave


def dragY1(t: Union[int, float], a: float, tau: float, sigma: float, t0: Union[int, float] = 0.,
           freq: float = None, phase: float = None, phase0: float = None) -> QWaveform:
    """
    Return a QWaveform object of DRAG wave.

    :param t: pulse duration
    :param a: pulse amplitude
    :param tau: pulse center position
    :param sigma: pulse standard deviation
    :param t0: start time
    :param freq: pulse frequency shift (will not accumulate during the entire pulse execution)
    :param phase: pulse phase shift (will accumulate during the entire pulse execution)
    :param phase0: pulse phase shift (will not accumulate during the entire pulse execution)
    :return: returned QWaveform object with DRAG
    """

    def func(_t, args):
        """ The DRAG function """
        _a, _tau, _sigma = args
        if sigma == 0:
            return 0
        pulse = - _a * (_t - _tau) / (_sigma ** 2) * exp(- ((_t - _tau) / _sigma) ** 2 / 2)
        return pulse

    wave = QWaveform(f=func, t0=t0, t=t, args=(a, tau, sigma), freq=freq, phase=phase, phase0=phase0)
    wave.name = "drag_y1"
    return wave


def mix(xWave: Optional[QWaveform], yWave: Optional[QWaveform], t0: Union[int, float] = 0.):
    """
    Return the mix of two QWaveform instances.

    :param xWave: the first QWaveform instances
    :param yWave: the second QWaveform instances
    :param t0:
    :return: returned mixed QWaveform
    """

    def func(_t, args):
        """ The mix function """
        _xWave, _yWave = args
        _xVal = 0. if _xWave is None else _xWave(_t)
        _yVal = 0. if _yWave is None else _yWave(_t)
        # Calculate amplitude
        amp = math.sqrt(_xVal ** 2 + _yVal ** 2)
        # Calculate phase
        if abs(_xVal) > 1e-10 and abs(_yVal) > 1e-10:
            varPhase = atan(_yVal / _xVal)
        elif abs(_xVal) < 1e-10 < abs(_yVal):
            varPhase = sign(_yVal) * pi / 2
        else:
            varPhase = 0.
        return amp, varPhase

    xDuration = 0. if xWave is None else xWave.t0 + xWave.t
    yDuration = 0. if yWave is None else yWave.t0 + yWave.t
    tDuration = max(xDuration, yDuration)
    wave = QWaveform(f=func, t0=t0, t=tDuration, args=(xWave, yWave), flag=QWAVEFORM_FLAG_MIX_WAVE)
    xName = "" if xWave is None else xWave.name
    yName = "" if yWave is None else yWave.name
    wave.name = "mix-{}-{}".format(xName, yName)
    return wave



#!/usr/bin/python3
# -*- coding: utf8 -*-

# Copyright (c) 2021 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Plot functions.
"""

import numpy
import matplotlib.pyplot as plt
from matplotlib.ticker import FormatStrFormatter
from typing import List, Dict, Any, Union
import matplotlib.ticker as mtick
from Quanlse.QPlatform.Error import ArgumentError


def plotBarGraph(x: List[Any], y: List[float], title: str = "", xLabel: str = "", yLabel: str = "",
                 color: str = 'blue', lineWidth: float = 1.0, fontSize: float = 10, spacing: float = 0.1) \
        -> None:
    """
    Plot bar graphs with given X labels and Y values lists, this function also supports other optional
    parameters as shown below.

    :param x: a list of X labels for the graphs
    :param y: a list of Y coordinates for the bar graphs
    :param title: overall title of the graph
    :param xLabel: label on X axis
    :param yLabel: label on Y axis
    :param color: a color for all bars, default at blue (from mint, blue, red, green, yellow, black, pink, cyan,
        purple, darkred, orange, brown, pink and teal)
    :param lineWidth: line width index
    :param fontSize: font size index
    :param spacing: vertical/horizontal spacing index
    :return: None
    """
    # define color map
    cMap = colorMap()

    xList = numpy.arange(len(y))

    # Create figure instance
    plt.figure(1, (15, 10))

    # Set graph size and background color
    plt.rcParams["figure.figsize"] = (15, 10)
    plt.rcParams['axes.linewidth'] = lineWidth * 2

    # title and size
    if title is not None:
        plt.title(title, size=int(fontSize * 1.4), weight='bold')

    # label and size
    if xLabel != "":
        plt.xlabel(xLabel, size=int(fontSize * 1.2), weight='bold')
    if yLabel != "":
        plt.ylabel(yLabel, size=int(fontSize * 1.2), weight='bold')

    # Ticks size
    plt.xticks(fontsize=fontSize, weight='bold')
    plt.yticks(fontsize=fontSize, weight='bold')

    # adjust vertical and horizontal spacing
    plt.xlim([min(xList) - (max(xList) - min(xList)) * spacing, max(xList) + (max(xList) - min(xList)) * spacing])
    if min(y) >= 0:
        plt.ylim([min(y), max(y) + (max(y) - min(y)) * spacing])
    else:
        plt.ylim([min(y) - (max(y) - min(y)) * spacing, max(y) + (max(y) - min(y)) * spacing])

    # configure color and plot
    if color is None:
        plt.bar(xList, y, alpha=0.7, linewidth=lineWidth)
    else:
        for aColor in cMap:
            if aColor[0] == color:
                plt.bar(xList, y, color=aColor[1], alpha=1, linewidth=lineWidth)
    plt.xticks(xList, x)

    # draw the line of y = 0
    plt.axhline(y=0, color='#ccc', linestyle=':', linewidth=lineWidth)
    plt.show()


def plotLineGraph(x: Union[List[List[float]], List[float]], y: Union[List[List[float]], List[float]], title: str = "",
                  xLabel: str = "", yLabel: str = "",
                  legends: List[str] = None, color: List[str] = None,
                  lineWidth: float = 2.0,
                  fontSize: float = 10, spacing: float = 0.05, log=False) \
        -> None:
    """
    Plot line graph, with the given 2-dimensional lists of X and Y values,
    this function also supports other optional parameters as shown below.

    :param x: a 2-dim list of X coordinates for the line graphs
    :param y: a 2-dim list of Y coordinates for the line graphs
    :param title: overall title of the graph
    :param xLabel: label on X axis
    :param yLabel: label on Y axis
    :param legends: a list of strings - a label each string
    :param color: a list of colors - a color for each line (unspecified lines will be of default color)
        (from mint, blue, red, green, yellow, black, pink, cyan, purple, darkred, orange, brown, pink and teal)
    :param lineWidth: line width index
    :param fontSize: font size index
    :param spacing: vertical spacing index
    :param log: log, default at false
    :return: None
    """

    # Default color theme
    if color is None:
        color = ['blue', 'pink', 'cyan', 'mint', 'red']

    # take log
    if log:
        tmp = y
        y = numpy.log(numpy.array(y))
        if y.min() == -numpy.Infinity:
            y = tmp
            print("can't take log of zero, switch back to non-log mode")
    # Define Color Map
    cMap = colorMap()

    # Set graph size and background color
    plt.rcParams["figure.figsize"] = (8, 6)
    plt.rcParams['axes.linewidth'] = lineWidth * 0.75

    # Check the type of the input of x and y
    if not isinstance(x[0], list):
        x = [x]
        y = [y]

    # Create figure instance
    plt.figure(1, (15, 10))

    # find max and min for spacing purposes
    maximumX = max(x[0])
    minimumX = min(x[0])
    maximumY = max(y[0])
    minimumY = min(y[0])

    # plot and look for global min/max
    for i in range(0, len(x)):
        if max(y[i]) > maximumY:
            maximumY = max(y[i])
        if min(y[i]) < minimumY:
            minimumY = min(y[i])
        if max(x[i]) > maximumX:
            maximumX = max(x[i])
        if min(x[i]) < minimumX:
            minimumX = min(x[i])

        if (color is None) or (i < 0) or (i >= len(color)):
            if legends is not None:
                plt.plot(x[i], y[i], label=legends[i], linewidth=lineWidth, alpha=1)
            else:
                plt.plot(x[i], y[i], linewidth=lineWidth)
        else:
            flag = False
            for aColor in cMap:
                if aColor[0] == color[i]:
                    flag = True
                    if legends is not None:
                        plt.plot(x[i], y[i], color=aColor[1], linewidth=lineWidth, label=legends[i], alpha=1)
                    else:
                        plt.plot(x[i], y[i], color=aColor[1], linewidth=lineWidth, alpha=1)
            if flag is False:
                raise ValueError('Color not found!')

    # spacing
    plt.xlim([minimumX - (maximumX - minimumX) * spacing, maximumX + (maximumX - minimumY) * spacing])
    plt.ylim([minimumY - (maximumY - minimumY) * spacing, maximumY + (maximumY - minimumY) * spacing])

    # Ticks size
    plt.xticks(fontsize=fontSize, weight='bold')
    plt.yticks(fontsize=fontSize, weight='bold')

    # title and size
    if title is not None:
        plt.title(title, size=int(fontSize * 1.9), weight='bold')

    # label and size
    if xLabel is not None:
        plt.xlabel(xLabel, size=int(fontSize * 1.4), weight='bold')
    if yLabel is not None:
        plt.ylabel(yLabel, size=int(fontSize * 1.4), weight='bold')

    plt.legend()
    plt.show()


def plotPulse(x: Union[List[List[float]], List[float]],
              y: Union[List[List[Union[float, complex]]], List[Union[float, complex]]],
              title: Union[List[str], str], yLabel: Union[List[str], str], color: Union[List[str], str],
              xLabel: str = '', lineWidth: float = 1.0, fontSize: float = 10, spacing: float = 0.1, dark: bool = False,
              complx: bool = False) -> None:
    """
    Plots pulse graphs, and is used in the plotWave() function in the Hamiltonian file. This function takes
    two mandatory two-dimensional lists of X and Y values. This function also supports other optional parameters
    as shown below.

    :param x: a 2-dim list consisting the time coordinates for each pulse
    :param y: a 2-dim list consisting the amplitudes for each pulse
    :param title: a list of titles, one for each pulse
    :param yLabel: a lists of y labels for each pulse
    :param xLabel: a string of x label (for all pulse)
    :param color: a list of colors for each pulse
        (from mint, blue, red, green, yellow, black, pink, cyan, purple, darkred, orange, brown, pink and teal)
        If pulse consists more than 250 cuts, switch to slice
    :param lineWidth: line width index
    :param fontSize: font size index
    :param spacing: vertical spacing index
    :param dark: enables a dark-themed mode
    :param complx: enable complex plotting
    :return: None
    """

    # Set color theme
    if dark:
        fontColor = 'ghostwhite'
        bgColor = '#1B1C21'
        axisColor = 'grey'
    else:
        fontColor = 'black'
        bgColor = 'white'
        axisColor = 'black'

    # Import colormap
    cMap = colorMap()

    # create fig and ax and subplots to plot
    if len(x) == 1:
        fig, ax1 = plt.subplots(len(x), 1, figsize=(16, len(x) * 1.5 + 3/4))
        ax = [ax1]
    else:
        fig, ax = plt.subplots(len(x), 1, figsize=(16, len(x) * 1.5 + 3/4))
    fig.patch.set_facecolor(bgColor)

    # loop through the list of graphs
    for k in range(0, len(x)):

        # if lengths of the two lists don't match
        if len(x[k]) != len(y[k]):
            raise ArgumentError("Lengths of lists do not match")

        # f list x contains more than 250 slices, switch to line plot
        maxPieces = 250
        plotType = 'slice'
        if len(y[k]) > maxPieces or complx is True:
            plotType = 'line'

        # check whether color exists in colormap
        colorFound = False

        # find according color in colormap
        for aColor in cMap:

            # if color found
            if aColor[0] == color[k]:
                colorFound = True

                # plot graph (line/bar)
                if plotType == 'line':
                    realY = [y[k][i].real for i in range(len(y[k]))]
                    comY = [y[k][i].imag for i in range(len(y[k]))]
                    ax[k].axhline(linewidth=lineWidth * 1.3, color=axisColor, alpha=0.4, ls=':')
                    ax[k].plot(x[k], realY, color=aColor[1], linewidth=lineWidth * 1.3)
                    ax[k].fill_between(x[k], 0, realY, facecolor=aColor[1], alpha=0.07)

                    if complx:
                        ax[k].plot(x[k], comY, color='darkviolet', linewidth=lineWidth * 1.3)
                        ax[k].fill_between(x[k], 0, comY, facecolor='darkviolet', alpha=0.05)
                elif plotType != 'line':
                    # shift pulse so that first pulse starts on 0
                    for i in x[k]:
                        i += (x[k][1] - x[k][0]) / 2
                    ax[k].bar(x[k], y[k], color=bgColor, alpha=1,
                              width=x[k][1] - x[k][0], edgecolor=aColor[1], linewidth=lineWidth * 1.3)
                    ax[k].bar(x[k], y[k], color=aColor[1], alpha=0.05,
                              width=x[k][1] - x[k][0], linewidth=0)
                break

        # if invalid color
        if not colorFound:
            raise ValueError('Color not found!')

        # adjust spacing between borders
        if complx is False:
            if type(y[k][0]) is complex:
                raise ValueError('Complex value found in non-complex mode!')
            if max(y[k]) == min(y[k]):
                ax[k].set_ylim(min(y[k]) - (spacing * 10), max(y[k]) + (spacing * 10))
            else:
                ax[k].set_ylim(
                    [min(y[k]) - (max(y[k]) - min(y[k])) * spacing * 2.5,
                     max(y[k]) + (max(y[k]) - min(y[k])) * (spacing + 0.3)])
            ax[k].set_xlim(min(x[k]), max(x[k]))
        else:
            # realParts = [y[k]]
            maxY = max(max(realY), max(comY))
            minY = min(min(realY), min(comY))
            if maxY == minY:
                ax[k].set_ylim(minY - (spacing * 10), maxY + (spacing * 10))
            else:
                ax[k].set_ylim(
                    [minY - (maxY - minY) * spacing * 2.5,
                     maxY + (maxY - minY) * (spacing + 0.3)])
            ax[k].set_xlim(min(x[k]), max(x[k]))
        # if last, add label, otherwise remove ticks,
        if k != len(x) - 1:
            ax[k].set_xticks([])
        else:
            ax[k].set_xlabel(xLabel, size=fontSize * 1.6)

        # set title for each graph
        ax[k].set_title(title[k] + "  ", fontdict={'fontsize': 15, 'fontweight': 'bold'},
                        y=0.75, loc='right', color=fontColor)

        # set y label
        ax[k].set_ylabel(yLabel[k], size=fontSize * 1.4)

        # set yTick decimal place
        ax[k].yaxis.set_major_formatter(FormatStrFormatter('%.1f'))

        # set tick size
        ax[k].tick_params(labelsize=fontSize * 1.3)

        # set spines linewidth
        # if k == 0:
        ax[k].spines["top"].set_linewidth(1.5)
        ax[k].spines["bottom"].set_linewidth(1.5)
        ax[k].spines["left"].set_linewidth(1.5)
        ax[k].spines["right"].set_linewidth(1.5)

        # set spines color
        ax[k].spines["bottom"].set_color(axisColor)
        ax[k].spines["right"].set_color(axisColor)
        ax[k].spines["left"].set_color(axisColor)
        ax[k].spines["top"].set_color(axisColor)

        # set label and ticks color
        ax[k].xaxis.label.set_color(fontColor)
        ax[k].yaxis.label.set_color(fontColor)
        ax[k].tick_params(axis='x', colors=fontColor)
        ax[k].tick_params(axis='y', colors=fontColor)

        # set facecolor
        ax[k].set_facecolor(bgColor)

        # check if scientific notation is needed
        fig.canvas.draw()
        for i in range(len(ax[k].get_yticklabels())):
            for j in range(len(ax[k].get_yticklabels())):
                if ax[k].get_yticklabels()[i].get_text() =='0.0' and ax[k].get_yticklabels()[j].get_text() =='-0.0':
                    ax[k].yaxis.set_major_formatter(mtick.FormatStrFormatter('%.2e'))
                    break
                if ax[k].get_yticklabels()[i].get_text() == ax[k].get_yticklabels()[j].get_text() and i != j:
                    ax[k].yaxis.set_major_formatter(mtick.FormatStrFormatter('%.2e'))
                    break

    # align y labels
    fig.align_ylabels(ax[:])
    plt.show()


def plotProcess(chi: Dict[str, Any]) -> None:
    """
    Generate the Process Tomography Plot in 3D. This function takes a mandatory dictionary Chi,
    which is directly generated from processTomography.

    :param chi: Chi Dictionary that is obtained from processTomography
    :return: None
    """

    # setup the figure and axes
    fig = plt.figure(figsize=(12, 6))
    ax2 = fig.add_subplot(121, projection='3d')
    ax1 = fig.add_subplot(122, projection='3d')

    numX = len(chi['XLabel'])
    numY = len(chi['YLabel'])

    # fake data
    _x = numpy.arange(numX)
    _y = numpy.arange(numY)
    _xx, _yy = numpy.meshgrid(_x, _y)
    x, y = _xx.ravel(), _yy.ravel()

    top = numpy.reshape(chi['Imag'], numX * numY, )
    bottom = numpy.zeros_like(top)
    width = depth = 0.7

    # plot graph
    ax1.bar3d(x, y, bottom, width, depth, top, shade=True, color='r')
    ax1.set_xticks(numpy.linspace(0.5, numX - 0.5, numX))
    ax1.set_yticks(numpy.linspace(0.5, numY - 0.5, numY))
    ax1.w_xaxis.set_ticklabels(chi['XLabel'], fontsize=12)
    ax1.w_yaxis.set_ticklabels(chi['YLabel'], fontsize=12)
    ax1.set_title('Chi Matrix - Imaginary Parts', fontsize=18)
    ax1.set_zlim(-0.5, 0.5)
    top = numpy.reshape(chi['Real'], numX * numY, )
    bottom = numpy.zeros_like(top)

    ax2.bar3d(x, y, bottom, width, depth, top, shade=True, color='c')
    ax2.set_xticks(numpy.linspace(0.5, numX - 0.5, numX))
    ax2.set_yticks(numpy.linspace(0.5, numY - 0.5, numY))
    ax2.w_xaxis.set_ticklabels(chi['XLabel'], fontsize=12)
    ax2.w_yaxis.set_ticklabels(chi['YLabel'], fontsize=12)
    ax2.set_title('Chi Matrix - Real Parts', fontsize=18)
    ax2.set_zlim(-0.5, 0.5)
    plt.show()


def plotHeatMap(matrix: numpy.ndarray, xTicks: List = None, yTicks: List = None, xLabel: str = "", yLabel: str = "",
                useLog: bool = False, cMap: str = "cividis") -> None:
    """
    Plot a 2D heat map.

    :param matrix: the values of each square
    :param xTicks: ticks on X axis
    :param yTicks: ticks on Y axis
    :param xLabel: label on X axis
    :param yLabel: label on Y axis
    :param useLog: format ticks on a log scale
    :param cMap: indicate color_map
    :return: None
    """

    ax = plt.gca()

    if useLog:
        _matrix = numpy.log10(matrix)
        im = ax.imshow(_matrix, cmap=cMap, vmin=-10, vmax=0)
    else:
        _matrix = matrix
        im = ax.imshow(_matrix, cmap=cMap)

    for edge, spine in ax.spines.items():
        spine.set_visible(False)

    ax.set_xticks(numpy.arange(numpy.shape(_matrix)[1]))
    ax.set_yticks(numpy.arange(numpy.shape(_matrix)[0]))
    ax.grid(which="minor", color="w", linestyle='-', linewidth=3)
    ax.tick_params(which="minor", bottom=False, left=False)

    if xTicks is None:
        ax.set_xticklabels(numpy.arange(numpy.shape(_matrix)[1]))
    else:
        ax.set_xticklabels(xTicks)

    if yTicks is None:
        ax.set_yticklabels(numpy.arange(numpy.shape(_matrix)[0]))
    else:
        ax.set_yticklabels(yTicks)

    im_ratio = numpy.shape(_matrix)[0] / numpy.shape(_matrix)[1]
    if useLog:
        cbar = ax.figure.colorbar(im, ax=ax, fraction=0.046 * im_ratio, pad=0.04)
        cbar.ax.set_ylabel(r"$\log_{10}(Population)$", rotation=-90, va="bottom")
    else:
        cbar = ax.figure.colorbar(im, ax=ax, fraction=0.046 * im_ratio, pad=0.04)
        cbar.ax.set_ylabel("Population", rotation=-90, va="bottom")

    valFormat = "{x:.2f}"
    threshold = im.norm(numpy.max(_matrix)) / 2.
    textColors = ("black", "white")
    kw = dict(horizontalalignment="center",
              verticalalignment="center")

    for i in range(numpy.shape(_matrix)[0]):
        for j in range(numpy.shape(_matrix)[1]):
            value = float(_matrix[i, j])
            kw.update(color=textColors[int(im.norm(value) < threshold)])
            im.axes.text(j, i, str(round(value, 2)), **kw)

    plt.xlabel(xLabel)
    plt.ylabel(yLabel)
    plt.show()


def colorMap():
    """
    Default color setting.

    :return: Color map
    """
    return [
        ['mint', 'mediumspringgreen'],
        ['lightblue', 'deepskyblue'],
        ['blue', 'cornflowerblue'],
        ['yellow', 'yellow'],
        ['green', 'forestgreen'],
        ['red', 'crimson'],
        ['gold', 'gold'],
        ['black', 'black'],
        ['cyan', 'darkturquoise'],
        ['purple', 'darkviolet'],
        ['teal', 'teal'],
        ['darkred', 'darkred'],
        ['orange', 'orangered'],
        ['brown', 'peru'],
        ['pink', 'deeppink']

    ]


def plotIonPosition(ionPos: List[Any]) -> None:
    """
    Plot the position of the ion in the equilibrium.

    :param ionPos: the positions of the ions.
    :return: None
    """
    plt.figure()
    plt.scatter(numpy.array(ionPos), numpy.zeros(numpy.size(ionPos)), s=100)
    plt.title('Ions equilibrium position', fontsize='large', fontweight='bold')

    font2 = {'family': 'Times New Roman',
             'weight': 'bold',
             'size': 15,
             }
    plt.xlabel(r'Z axial ($\mu$m)', font2)
    plt.ylabel('XY plane', font2)
    ax = plt.gca()
    ax.spines['bottom'].set_linewidth(2)
    ax.spines['left'].set_linewidth(2)
    ax.spines['right'].set_linewidth(2)
    ax.spines['top'].set_linewidth(2)
    plt.tick_params(labelsize=15)
    plt.show()


def plotPop(x: List[Any], y: List[float], title: str = "", xLabel: str = "", yLabel: str = "",
                 color: str = 'purple', lineWidth: float = 1.0, fontSize: float = 10, spacing: float = 0.1) \
        -> None:
    """
    Plot bar graphs with given X labels and Y values lists, this function also supports other optional
    parameters as shown below.

    :param x: a list of X labels for the graphs
    :param y: a list of Y coordinates for the bar graphs
    :param title: overall title of the graph
    :param xLabel: label on X axis
    :param yLabel: label on Y axis
    :param color: a color for all bars, default at blue (from mint, blue, red, green, yellow, black, pink, cyan,
        purple, darkred, orange, brown, pink and teal)
    :param lineWidth: line width index
    :param fontSize: font size index
    :param spacing: vertical/horizontal spacing index
    :return: None
    """
    # define color map
    cMap = colorMap()

    xList = numpy.arange(len(y))

    # Create figure instance
    plt.figure(1, (15, 10))

    # Set graph size and background color
    plt.rcParams["figure.figsize"] = (15, 10)
    plt.rcParams['axes.linewidth'] = lineWidth * 2

    # title and size
    if title is not None:
        plt.title(title, size=int(fontSize * 1.4), weight='bold')

    # label and size
    if xLabel != "":
        plt.xlabel(xLabel, size=int(fontSize * 1.2), weight='bold')
    if yLabel != "":
        plt.ylabel(yLabel, size=int(fontSize * 1.2), weight='bold')

    # Ticks size
    plt.xticks(fontsize=fontSize, weight='bold')
    plt.yticks(fontsize=fontSize, weight='bold')

    # adjust vertical and horizontal spacing
    plt.xlim([min(xList) - (max(xList) - min(xList)) * spacing, max(xList) + (max(xList) - min(xList)) * spacing])
    if min(y) >= 0:
        plt.ylim([0, 1])
    else:
        plt.ylim([min(y) - (max(y) - min(y)) * spacing, max(y) + (max(y) - min(y)) * spacing])

    # configure color and plot
    if color is None:
        plt.bar(xList, y, alpha=0.7, linewidth=lineWidth)
    else:
        for aColor in cMap:
            if aColor[0] == color:
                plt.bar(xList, y, color=aColor[1], alpha=1, linewidth=lineWidth)
    plt.xticks(xList, x)

    # draw the line of y = 0
    plt.axhline(y=0, color='#ccc', linestyle=':', linewidth=lineWidth)
    plt.show()



#!/usr/bin/python3
# -*- coding: utf8 -*-

# Copyright (c) 2020 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Simulating the time evolution of the qubits at the pulse level gives us more insight into how
quantum gates operates and the effects of noise. For superconducting quantum circuit, the
transmon qubits are controlled by applying microwave and magnetic flux. However, the performance
of quantum gates are often suppressed by various factors - the decoherence of the qubit due
to its interaction with the environment, the unwanted cross-talk effect and leakage into the
higher levels of the transmon.

The pulse-level simulator provided by Quanlse allows us to simulate quantum operations on
noisy quantum device consisting of multiple transmon qubits to better understand the physics
behind quantum computing.

For more details, please visit https://quanlse.baidu.com/#/doc/tutorial-pulse-level-simulator.
"""

import random
import copy
import numpy
from typing import List, Union, Optional, Dict, Tuple, Callable, Any
from numpy import pi, sqrt, ndarray, zeros, identity, random, cos, sin, trace, round

from Quanlse.QHamiltonian import QHamiltonian
from Quanlse.QPlatform import Error
from Quanlse.QOperator import QOperator, duff, create, destroy, a, adagger, number, driveX, uWave
from Quanlse.QWaveform import QJob, square, QWaveform, QJobList, QResult, sequence, \
    QWAVEFORM_FLAG_ALWAYS_ON, QWAVEFORM_FLAG_DO_NOT_CLEAR, QWAVEFORM_FLAG_DO_NOT_PLOT
from Quanlse.Utils.Functions import tensor, basis, dagger
from Quanlse.Scheduler.Superconduct import SchedulerSuperconduct, SchedulerPulseGenerator
from Quanlse.QPlatform.Error import ArgumentError


class PulseModel(SchedulerSuperconduct):
    """
    Basic class of PulseModel. It defines the system information and noise-related parameters. It is also a sub-class of
    SchedulerSuperconduct.

    :param subSysNum: size of the subsystem
    :param sysLevel: the energy levels of the system (support different levels for different qubits)
    :param qubitFreq: qubit frequencies, in 2 pi GHz
    :param driveFreq: drive frequencies, in 2 pi GHz
    :param qubitAnharm: Anharmonicity for each qubit
    :param dt: sampling time period:
    :param couplingMap: direct coupling structure
    :param T1: T1-relaxation time, in nanoseconds
    :param T2: T2-dephasing time, in nanoseconds
    :param ampSigma: deviation of the amplitude distortion
    :param betaStd: standard deviation of the relative strength in classical crosstalk
    :param frameMode: rotating frame (in qubit frequency)
    :param pulseGenerator: customized pulse generator for scheduler
    """

    def __init__(self,
                 subSysNum: int,
                 sysLevel: Union[int, List[int]],
                 qubitFreq: Dict[int, Union[int, float]],
                 qubitAnharm: Dict[int, Union[int, float]],
                 dt: float = 0.2,
                 driveFreq: Dict[int, Union[int, float]] = None,
                 couplingMap: Optional[Dict[Tuple, Union[int, float]]] = None,
                 T1: Optional[Dict[int, Union[int, float, None]]] = None,
                 T2: Optional[Dict[int, Union[int, float, None]]] = None,
                 ampSigma: Optional[Union[int, float]] = None,
                 betaStd: Optional[Union[int, float]] = None,
                 frameMode: str = 'rot',
                 pulseGenerator: Callable = None):

        self.frameMode = frameMode  # type: str

        self._subSysNum = subSysNum  # type: int
        self._sysLevel = sysLevel  # type: int
        self._dt = dt  # type: float
        self._conf = {}  # type: Dict

        self._qubitFreq = qubitFreq  # type: Dict[int, Union[int, float]]
        self._driveFreq = {} if driveFreq is None else driveFreq  # type: Dict[int, Union[int, float]]
        self._qubitAnharm = qubitAnharm  # type: Dict[int, Union[int, float]]
        self._T1 = T1  # type: Optional[Dict[int, Union[int, float]]]
        self._T2 = T2  # type: Optional[Dict[int, Union[int, float]]]
        self._couplingMap = couplingMap  # type: Optional[Dict[Tuple, Union[int, float]]]
        self._pulseGenerator = None  # type: Optional[SchedulerPulseGenerator]
        self._pulseGeneratorFunc = pulseGenerator  # type: Callable

        super(PulseModel, self).__init__(dt, ham=None, subSysNum=subSysNum,
                                         sysLevel=sysLevel, generator=None)

        self.createQHamiltonian(frameMode=self.frameMode)

        self.betaStd = betaStd  # type: Optional[Union[int, float]]
        self.ampSigma = ampSigma  # type: Optional[Union[int, float]]

        if self._pulseGeneratorFunc is not None:
            self._pulseGenerator = self._pulseGeneratorFunc(self.ham)

    @property
    def dim(self) -> int:
        """
        Return the dimension of the Hilbert space.
        """
        _dim = 1
        if isinstance(self.sysLevel, int):
            _dim = self.sysLevel ** self.subSysNum

        elif isinstance(self.sysLevel, list):
            _dim = 1
            for level in self.sysLevel:
                _dim = _dim * level

        return _dim

    @property
    def qubitFreq(self) -> Dict[int, Union[int, float]]:
        """
        Return the eigenfrequency of each qubit in the system.
        """
        return self._qubitFreq

    @qubitFreq.setter
    def qubitFreq(self, value: Dict[int, Union[int, float]]):
        """
        Return the eigen frequency of each qubit in the system, in 2 * pi * GHz.
        """
        if len(value) != self.subSysNum:
            raise Error.ArgumentError("The number of the frequencies and the number of qubits are not the same.")

        self._qubitFreq = value
        self.createQHamiltonian(frameMode=self.frameMode)

    @property
    def driveFreq(self) -> Dict[int, Union[int, float]]:
        """
        Return the drive frequency of each qubit in the system.
        """
        return self._qubitFreq

    @driveFreq.setter
    def driveFreq(self, value: Dict[int, Union[int, float]]):
        """
        Return the drive frequency of each qubit in the system, in 2 * pi * GHz.
        """
        if len(value) != self.subSysNum:
            raise Error.ArgumentError("The number of the frequencies and the number of qubits are not the same.")

        self._driveFreq = value
        self.createQHamiltonian(frameMode=self.frameMode)

    @property
    def qubitAnharm(self) -> Dict[int, Union[int, float]]:
        """
        Return the anharmonicity of each qubit in the system.
        """
        return self._qubitAnharm

    @qubitAnharm.setter
    def qubitAnharm(self, value: Dict[int, Union[int, float]]):
        """
        Setter for qubits anharmonicity.
        """
        if len(value) != self.subSysNum:
            raise Error.ArgumentError("The number of the anharmonicities and the number of qubits are not the same.")

        self._qubitAnharm = value
        self.createQHamiltonian(frameMode=self.frameMode)

    @property
    def couplingMap(self) -> Optional[Dict[Tuple, Union[int, float]]]:
        """
        Return the eigenfrequency of each qubit in the system.
        """
        return self._couplingMap

    @couplingMap.setter
    def couplingMap(self, value: Optional[Dict[Tuple, Union[int, float]]]):
        """
        Return the coupling structure of the system.
        """
        if self.couplingMap is not None:
            indexList = [index for pair in value for index in pair]
            for index in indexList:
                if index not in range(self.subSysNum+1):
                    raise Error.ArgumentError("The number of the frequencies and the number of qubits are not the same.")

        self._couplingMap = value
        self.createQHamiltonian(frameMode=self.frameMode)

    @property
    def T1(self) -> Optional[Dict[int, Union[int, float, None]]]:
        """
        Return the relaxation time T1 for each qubit.
        """
        return self._T1

    @T1.setter
    def T1(self, value: Optional[Dict[int, Union[int, float, None]]]):
        """
        Setter for the relaxation time T1 for each qubit.
        """
        if value is not None:
            for index, t1 in value.items():
                if t1 < 0:
                    raise Error.ArgumentError("The value of T1 should be a positive real number.")
                if index not in range(self.subSysNum):
                    raise Error.ArgumentError(f"{index}th-qubit is not defined.")

        self._T1 = value
        self.createQHamiltonian(frameMode=self.frameMode)

    @property
    def T2(self) -> Optional[Dict[int, Union[int, float, None]]]:
        """
        Return the relaxation time T2 for each qubit.
        """
        return self._T2

    @T2.setter
    def T2(self, value: Optional[Dict[int, Union[int, float, None]]]):
        """
        Setter for the relaxation time T2 for each qubit.
        """
        if value is not None:
            for index, t2 in value.items():
                if t2 < 0:
                    raise Error.ArgumentError("The value of T1 should be a positive real number.")
                if index not in range(self.subSysNum):
                    raise Error.ArgumentError(f"{index}th-qubit is not defined.")

        self._T2 = value
        self.createQHamiltonian(frameMode=self.frameMode)

    def diagMat(self) -> ndarray:
        """
        Return the diagonal matrix of the Hamiltonian.

        :return: a diagonal matrix
        """

        levelList = []
        identityList = []

        # Define the list of different identity matrix for each subsystem
        if isinstance(self.sysLevel, int):
            levelList = [self.sysLevel for _ in range(self.subSysNum)]
            identityList = [identity(self.sysLevel) for _ in range(self.subSysNum)]

        elif isinstance(self.sysLevel, list):
            levelList = self.sysLevel
            identityList = [identity(self.sysLevel[i]) for i in range(self.subSysNum)]

        # Initialize the diagonal term
        diagTerm = zeros([self.dim, self.dim], dtype=complex)

        # Add drift terms to the diagonal term
        for index, wq in self.qubitFreq.items():
            numMat = copy.deepcopy(identityList)
            numMat[index] = numMat[index] @ number(levelList[index]).matrix
            diagTerm += wq * tensor(numMat)

        return diagTerm

    def _generateDrift(self, ham: QHamiltonian, frameMode: str = 'rot') -> None:
        """
        Generate the drift Hamiltonian.

        :param ham: The QHamiltonian object.
        :param frameMode: Hamiltonian in a rotating frame.
        :return: None
        """

        # Define the rotating frame of the system

        if frameMode == 'lab':

            # create drift term in the lab frame

            if isinstance(self.sysLevel, int):
                for index, freq in self.qubitFreq.items():
                    ham.addDrift(operator=number(self.sysLevel), coef=freq, onSubSys=index)
                for index, anharm in self.qubitAnharm.items():
                    ham.addDrift(operator=duff(self.sysLevel), coef=0.5 * anharm, onSubSys=index)
            elif isinstance(self.sysLevel, list):
                for index, freq in self.qubitFreq.items():
                    ham.addDrift(operator=number(self.sysLevel[index]), coef=freq, onSubSys=index)
                for index, anharm in self.qubitAnharm.items():
                    ham.addDrift(operator=duff(self.sysLevel[index]), coef=0.5 * anharm, onSubSys=index)

        elif frameMode == 'rot':

            # create drift term in the rotating frame

            if isinstance(self.sysLevel, int):

                for index, anharm in self.qubitAnharm.items():
                    ham.addDrift(operator=duff(self.sysLevel), coef=0.5 * anharm, onSubSys=index)

            if isinstance(self.sysLevel, list):
                for index, anharm in self.qubitAnharm.items():
                    ham.addDrift(operator=duff(self.sysLevel[index]), coef=0.5 * anharm, onSubSys=index)

        else:
            raise Error.ArgumentError(f"Unsupported frameMode {frameMode}, only 'lab' and 'rot' are supported.")

    def _generateTimeIndepCoupTerm(self, ham: QHamiltonian) -> None:
        r"""
        Generate time-independent coupling terms for ith qubit and jth qubit.
        :math:`H_{coup} = a_{i} a_{j}^\dagger + a_{i}^\dagger a_j`

        :param ham: The QHamiltonian object.
        :return: None
        """
        # Generate the time independent coupling terms
        for index, value in self.couplingMap.items():
            ham.addCoupling(onSubSys=[index[0], index[1]], g=value)

    def _generateCoupTerm(self, ham: QHamiltonian) -> None:
        r"""
        Return the time-dependent coupling terms of ith qubit and jth qubit.

        :math:`H_{coup} = (a_i^\dagger a_j + a_i a_j^\dagger) \cos(\omega_{qi} - \omega_{qj})t
            + (a_i^\dagger a_j - a_i a_j^\dagger) i \sin(\omega_{qi} - \omega_{qj})t`

        :param ham: a QHamiltonian object.
        :return: None
        """

        for index, value in self.couplingMap.items():

            # Calculate the detuning of qubit i and qubit j

            deltaOmega = self.qubitFreq[index[0]] - self.qubitFreq[index[1]]

            ai = QOperator('ai')
            aj = QOperator('aj')
            adagi = QOperator('adagi')
            adagj = QOperator('adagj')

            # Define the operators for qubit i and qubit j

            if isinstance(self.sysLevel, int):
                ai = destroy(self.sysLevel)
                adagi = create(self.sysLevel)
                aj = destroy(self.sysLevel)
                adagj = create(self.sysLevel)

            elif isinstance(self.sysLevel, list):
                ai = destroy(self.sysLevel[index[0]])
                adagi = create(self.sysLevel[index[0]])
                aj = destroy(self.sysLevel[index[1]])
                adagj = create(self.sysLevel[index[1]])

            # Add the time-dependent always-on interaction between qubit i and qubit j

            ham.addWave(operators=[adagi, aj], onSubSys=[index[0], index[1]],
                        waves=square(t=0, a=value), freq=deltaOmega, phase=0., tag='coupling', t0=0,
                        flag=QWAVEFORM_FLAG_ALWAYS_ON | QWAVEFORM_FLAG_DO_NOT_CLEAR | QWAVEFORM_FLAG_DO_NOT_PLOT)

            ham.addWave(operators=[ai, adagj], onSubSys=[index[0], index[1]],
                        waves=square(t=0, a=value), freq=deltaOmega, phase=0., tag='coupling', t0=0,
                        flag=QWAVEFORM_FLAG_ALWAYS_ON | QWAVEFORM_FLAG_DO_NOT_CLEAR | QWAVEFORM_FLAG_DO_NOT_PLOT)

            ham.addWave(operators=[QOperator(name='adagi', matrix=1j * adagi.matrix), aj],
                        onSubSys=[index[0], index[1]], waves=square(t=0, a=value),
                        freq=deltaOmega, phase=-pi / 2, tag='coupling', t0=0,
                        flag=QWAVEFORM_FLAG_ALWAYS_ON | QWAVEFORM_FLAG_DO_NOT_CLEAR | QWAVEFORM_FLAG_DO_NOT_PLOT)

            ham.addWave(operators=[QOperator(name='ai', matrix=-1j * ai.matrix), adagj],
                        onSubSys=[index[0], index[1]], waves=square(t=0, a=value),
                        freq=deltaOmega, phase=-pi / 2, tag='coupling', t0=0,
                        flag=QWAVEFORM_FLAG_ALWAYS_ON | QWAVEFORM_FLAG_DO_NOT_CLEAR | QWAVEFORM_FLAG_DO_NOT_PLOT)

    def _setDecoherence(self, ham: QHamiltonian) -> None:
        r"""
        Define the collapse operators accounts for the decoherence noise
        for solving Lindblad master equation of the open system evolution.

        :math:`C_{\rm relaxation} = \frac{1}{\sqrt{T1}} a`
        :math:`C_{\rm dephasing} = \frac{1}{\sqrt{T2}} a^\dagger a`

        :param ham: The QHamiltonian object.
        :return: None
        """
        levelList = []

        # Compute the list of different level for each qubit
        if isinstance(self.sysLevel, list):
            levelList = self.sysLevel

        elif isinstance(self.sysLevel, int):
            levelList = [self.sysLevel for _ in range(self.subSysNum)]

        if self.T1 is None and self.T2 is None:
            pass

        # Initialize the dict type data of T1 and T2 for each qubit
        else:
            t1List = {}
            t2List = {}

            for index in range(self.subSysNum):
                t1List[index] = None
                t2List[index] = None

            if self.T1 is not None:
                for index, t1 in self.T1.items():
                    t1List[index] = t1

            if self.T2 is not None:
                for index, t2 in self.T2.items():
                    t2List[index] = t2

            # Initialize the dict type data for the collapse operators
            cList = {}

            # Calculate the collapse operators for each qubit based on t1List and t2List

            for index in t1List.keys():
                t1 = t1List[index]
                t2 = t2List[index]

                if t1 is None and t2 is None:
                    continue

                elif t2 is not None:
                    cList[index] = []
                    if t1 is not None:
                        rate1 = 1 / t1
                        mat = a(levelList[index])
                        cList[index].append(sqrt(rate1) * mat)

                        rate2 = 0.5 * ((1 / t2) - (1 / (2 * t1)))
                        mat = 2 * adagger(levelList[index]) @ a(levelList[index])
                        cList[index].append(sqrt(rate2) * mat)
                    else:
                        rate2 = 0.5 * (1 / t2)
                        mat = 2 * adagger(levelList[index]) @ a(levelList[index])
                        cList[index].append(sqrt(rate2) * mat)

                elif t1 is not None:
                    cList[index] = []
                    rate1 = 1 / t1
                    mat = a(levelList[index])
                    cList[index].append(sqrt(rate1) * mat)

            ham.setCollapse(cList)

    def _addCrossTalk(self, job: QJob) -> None:
        r"""
        Add classical crosstalk to a QJob object.

        :param job: a QJob object.
        :return: None
        """

        # Create the cross talk matrix of relative drive strength
        std = self.betaStd  # Standard deviation of the relative drive strength
        betaMat = zeros([self.subSysNum, self.subSysNum])
        jobCopy = copy.deepcopy(job)

        # Add standard deviation to the matrix
        for index, value in self.couplingMap.items():
            if index[0] != index[1]:
                beta = random.normal(scale=std)
                betaMat[index[0]][index[1]] = beta
                betaMat[index[1]][index[0]] = beta

        # Create the cross talk matrix of phase lag

        phiMat = zeros([self.subSysNum, self.subSysNum])

        for index, value in self.couplingMap.items():
            if index[0] != index[1]:
                phi = random.random() * 2 * pi
                phiMat[index[0]][index[1]] = phi
                phiMat[index[1]][index[0]] = phi

        ctrlOperatorsDict = copy.deepcopy(jobCopy.ctrlOperators)

        # Add the classical crosstalk to the system

        for opKey, op in ctrlOperatorsDict.items():
            qubitsNum = self.subSysNum
            i = op.onSubSys
            for j in range(0, qubitsNum):
                if betaMat[i][j] != 0:
                    beta = betaMat[i][j]
                    phaseLag = phiMat[i][j]
                    waves = copy.deepcopy(jobCopy.waves[opKey])
                    if isinstance(waves, QWaveform):
                        waves = [waves]
                    for wave in waves:
                        if wave.phase is None:
                            phase = 0
                        else:
                            phase = copy.deepcopy(wave.phase)

                        if wave.freq is not None:
                            freq = copy.deepcopy(wave.freq)
                        else:
                            freq = 0

                        job.addWave(operators=op, onSubSys=j, waves=wave, strength=beta,
                                    freq=freq, phase=phase + phaseLag, tag='cross-talk',
                                    flag=QWAVEFORM_FLAG_DO_NOT_PLOT)

    def _addAmpNoise(self, job: QJob) -> None:
        r"""
        Add amplitude distortion to QJob object.

        :param job: a QJob object.
        :return: None
        """

        jobCopy = copy.deepcopy(job)
        jobCopy.clearWaves(tag='coupling')
        jobCopy.clearWaves(tag='cross-talk')
        jobCopy.clearWaves(tag='idle')
        jobCopy.buildWaveCache()

        for key, seq in jobCopy.waveCache.items():
            newSeq = []

            for amp in seq:
                noise = random.normal(scale=self.ampSigma)
                ampNoise = noise * amp  # Define the distortion to the pulse
                newSeq.append(ampNoise)

            job.addWave(name=key, waves=sequence(newSeq), t0=0, tag='amplitude-noise')

    def createQHamiltonian(self, frameMode: str = "rot") -> QHamiltonian:
        r"""
        Generate a QHamiltonian object based on the physics model and a QJob object.

        :param frameMode: the rotating frame we choose. The default setting is rotating frame being in the qubits
            frequencies.
        :return: a QHamiltonian object.
        """

        # Initialize the hamiltonian
        ham = QHamiltonian(subSysNum=self.subSysNum, sysLevel=self.sysLevel, dt=self.dt)

        # Create the system Hamiltonian in the lab frame
        if frameMode == "lab":
            # Create the system Hamiltonian in the lab frame
            self._generateDrift(ham, frameMode='lab')
            if self.couplingMap is not None:
                self._generateTimeIndepCoupTerm(ham)
            self._setDecoherence(ham)
        elif frameMode == "rot":
            # Create the system Hamiltonian in the rotating frame of qubit frequency
            self._generateDrift(ham)
            self._setDecoherence(ham)
            if self.couplingMap is not None:
                self._generateCoupTerm(ham)
        else:
            raise Error.ArgumentError(f"Unsupported frameMode {frameMode}, only 'lab' and 'rot' are supported.")

        self.ham = ham

        if self._driveFreq is not None:
            # Set the local oscillator
            for qIdx in self._driveFreq.keys():
                self.ham.job.setLO(uWave, qIdx, freq=self._driveFreq[qIdx])

        if self._pulseGeneratorFunc is not None:
            self._pulseGenerator = self._pulseGeneratorFunc(self.ham)

        return self.ham

    def getSimJob(self, job: QJob, noise: bool = True) -> QJob:
        r"""
        Add pulse-related noise to a QJob object.

        :param job: a QJob object.
        :param noise: Add noise to a job if true.

        :return: a QJob object.
        """

        # Check if the subSysNum and sysLevel agrees with the job

        if job.subSysNum != self.subSysNum:
            raise Error.ArgumentError("The variables 'subSysNum' are not matched.")

        if job.sysLevel != self.sysLevel:
            raise Error.ArgumentError("The variables 'sysLevel' are not matched.")

        simJob = copy.deepcopy(job)

        # Add crosstalk noise and amplitude distortion if noise is true

        if noise:
            if self.ampSigma is not None:
                self._addAmpNoise(simJob)

            if self.betaStd is not None:
                self._addCrossTalk(simJob)

        return simJob

    def simulate(self, job: QJob = None, state0: ndarray = None, jobList: QJobList = None,
                 measure: List[int] = None, shot: int = None, options: Any = None) -> QResult:
        """
        Calculate the unitary evolution operator with a given Hamiltonian. This function supports
        both single-job and batch-job processing.

        :param job: the QJob object to simulate.
        :param state0: the initial state vector. If None is given, this function will return the time-ordered
                       evolution operator, otherwise returns the final state vector.
        :param jobList: a job list containing the waveform list
        :param measure: indicates the qubit indexes for measurement
        :param shot: return the population of the eigenstates when ``shot`` is provided
        :param options: options pass to the runner function
        :return: result dictionary (or a list of result dictionaries when ``jobList`` is provided)

        This function does provide the option of simulating on local devices. However, Quanlse also provides
        cloud computing services which are significantly faster.

        **Example 1** (single-job processing):

        If ``jobList`` is None, the waveforms in ham will be used for simulation, and return a result dictionary.

        .. code-block:: python

                result = ham.simulate()

        **Example 2** (batch-job processing):

        If ``jobList`` is not None, simulation will use the waveform list for batch-job processing provided in
        jobList, and return a list of result dictionaries.

        .. code-block:: python

                result = ham.simulate(ham, jobList=jobs)

        Here, jobs is a QJobList object, indicating the the waveforms to be used.
        """

        if self.frameMode == 'lab':
            if state0 is None:
                state0 = basis(self.dim, 0)
            if shot is None:
                shot = 1000

        _useHamSimulator = True
        

        # Simulate the QJobs
        if _useHamSimulator:
            result = self._ham.simulate(job=job, state0=state0, jobList=jobList, shot=shot, adaptive=True)
        else:
            raise Error.RuntimeError('Not implemented')

        # Measure the population
        if measure is not None:
            if state0 is not None:
                pass
            else:
                raise ArgumentError("When 'measure' is given, you must input 'state0' or 'shot'!")

        return result


class ReadoutPulse:
    """
    A object contains the readout pulses to probe the resonator for qubit readout and demodulation.

    :param driveStrength: drive strength of the readout pulse.
    :param driveFreq: drive frequency of the readout pulse.
    :param loFreq: local oscillator's frequency for demodulation.
    """

    def __init__(self, driveStrength: Dict[int, Any], driveFreq: Dict[int, Any], loFreq: Union[int, float]):

        self._driveStrength = driveStrength
        self._driveFreq = driveFreq
        self._loFreq = loFreq

    @property
    def driveStrength(self) -> Dict[int, Any]:
        """
        Drive strength of readout pulse.
        """
        return self._driveStrength

    @driveStrength.setter
    def driveStrength(self, value: Dict[int, Any]):
        """
        Define the drive strength of readout pulse.
        """
        self._driveStrength = value

    @property
    def driveFreq(self) -> Dict[int, Any]:
        """
        Return the readout frequencies.
        """
        return self._driveFreq

    @driveFreq.setter
    def driveFreq(self, value: Dict[int, float]):
        """
        Define the drive frequencies of readout pulse.
        """
        self._driveFreq = value

    @property
    def loFreq(self):
        """
        Carrier frequency generated by the local oscillator of signal demodulation.
        """
        return self._loFreq

    @loFreq.setter
    def loFreq(self, value: float):
        """
        Set the carrier frequency generated by the local oscillator for signal demodulation.
        """
        self._loFreq = value

    def setFreq(self, index: int, freq: Union[int, float]):
        """
        Set the measurement channel's drive frequency of specified qubit.
        """
        self.driveFreq[index] = freq

    def setStrength(self, index: int, strength: Union[int, float]):
        """
        Set the strength of the measurement channel of specified qubit.
        """

        self.driveStrength[index] = strength


class ReadoutModel:
    """
    Basic class of readout simulation model, contains hardware information of the readout cavity.

    :param pulseModel: physics model of the simulator.
    :param resonatorFreq: bare frequencies of the resonators.
    :param level: energy levels of the resonators.
    :param coupling: qubit-resonator coupling strength.
    :param dissipation: dissipation parameter of the resonator.
    :param gain: the gain of the amplification.
    :param dt: sampling period of the AWG.
    :param impedance: impedance of the transmission line.
    :param conversionLoss: the conversion loss of the transmission line.
    """

    def __init__(self,
                 pulseModel: PulseModel,
                 resonatorFreq: Dict[int, float],
                 level: int,
                 coupling: Dict[int, float],
                 dissipation: Union[int, float],
                 gain: Union[int, float],
                 dt: float = 1.,
                 impedance: float = 1.,
                 conversionLoss: float = 1.):
        """
        The constructor of the readout model.
        """

        self.pulseModel = pulseModel  # type: PulseModel
        self.resonatorFreq = resonatorFreq  # type: Dict[int, float]
        self.dissipation = dissipation  # type:  Union[int, float]
        self.gain = gain  # type: Union[int, float]
        self.level = level  # type: int
        self.coupling = coupling  # type: Dict[int, float]
        self.dt = dt  # type: float
        self.impedance = impedance  # type: float
        self.conversionLoss = conversionLoss  # type: float

        self._readoutPulse = None  # type: Optional[ReadoutPulse]

    @property
    def resonatorFreq(self) -> Dict[int, float]:
        """
        Return the resonator frequencies.
        """

        return self._resonatorFreq

    @resonatorFreq.setter
    def resonatorFreq(self, value: Dict[int, float]):
        """
        Set the resonator frequencies.
        """
        qubitIdx = self.pulseModel.qubitFreq.keys()
        for resonatorIdx in value.keys():
            if resonatorIdx not in qubitIdx:
                raise Error.ArgumentError("Resonator indexes not found in qubit indexes")

            self._resonatorFreq = value

    @property
    def readoutPulse(self) -> ReadoutPulse:
        """
        The multiplexing pulse data for qubit readout.
        """
        return self._readoutPulse

    @readoutPulse.setter
    def readoutPulse(self, value: ReadoutPulse):
        """
        Set the measure channel for the readout
        """

        self._readoutPulse = value

    def _createHam(self, idx: int, driveFreq: Union[int, float]) -> QHamiltonian:
        """
        Return the hamiltonian of qubit-resonator circuit model according to the qubit index and drive frequency.

        :param idx: index of the qubit.
        :param driveFreq: drive frequency.

        :return: QHamiltonian in the rotating frame.
        """
        qubitFreq = self.pulseModel.qubitFreq
        qubitAnharm = self.pulseModel.qubitAnharm
        resonatorFreq = self.resonatorFreq
        qubitLevel = self.pulseModel.sysLevel
        cList = {1: [sqrt(self.dissipation) * a(self.level)]}

        ham = QHamiltonian(subSysNum=2, sysLevel=[qubitLevel, self.level], dt=self.dt)
        ham.addDrift(operator=number(qubitLevel), onSubSys=0, coef=qubitFreq[idx] - driveFreq)
        ham.addDrift(operator=duff(qubitLevel), onSubSys=0, coef=0.5 * qubitAnharm[idx])
        ham.addDrift(operator=number(self.level), onSubSys=1, coef=resonatorFreq[idx] - driveFreq)
        ham.addCoupling(onSubSys=[0, 1], g=self.coupling[idx])
        ham.setCollapse(cList)

        return ham

    def _createHamLab(self, idx: int) -> QHamiltonian:
        """
        Return the hamiltonian of qubit-resonator circuit model according to the qubit index and drive frequency.

        :param idx: index of the qubit.

        :return: QHamiltonian in the lab frame.
        """

        # Extract information from PulseModel object
        qubitFreq = self.pulseModel.qubitFreq
        qubitAnharm = self.pulseModel.qubitAnharm
        resonatorFreq = self.resonatorFreq
        qubitLevel = self.pulseModel.sysLevel
        cList = {1: [sqrt(self.dissipation) * a(self.level)]}

        # Create lab frame QHamiltonian object of a coupled transmon-resonator system
        ham = QHamiltonian(subSysNum=2, sysLevel=[qubitLevel, self.level], dt=self.dt)
        ham.addDrift(operator=number(qubitLevel), onSubSys=0, coef=qubitFreq[idx])
        ham.addDrift(operator=duff(qubitLevel), onSubSys=0, coef=0.5 * qubitAnharm[idx])
        ham.addDrift(operator=number(self.level), onSubSys=1, coef=resonatorFreq[idx])
        ham.addCoupling(onSubSys=[0, 1], g=self.coupling[idx])
        ham.setCollapse(cList)

        return ham

    def _readoutJob(self, amp: Union[int, float], duration: Union[int, float]) -> QJob:
        """
        Return readout pulse in rotating frame at driving frequency.

        :param amp: amplitude of the readout job.
        :param duration: the duration of the pulse.

        :return: QJob.
        """
        qubitLevel = self.pulseModel.sysLevel
        job = QJob(subSysNum=2, sysLevel=[qubitLevel, self.level], dt=self.dt)
        wave = square(t=duration, a=amp)
        job.addWave(operators=driveX, onSubSys=1, waves=wave, t0=0.)
        return job

    def simulate(self, duration: Union[int, float], resIdx: Optional[List[int]] = None,
                 state: 'str' = 'ground') -> Dict:
        """
        Simulate the state evolution of the qubit-resonator system using Jaynes-Cumming model.

        :param duration: pulse duration of the measurement.
        :param resIdx: indexes of the resonator the pulse acted on.
        :param state: qubit state prepared in ground state or in excited state, Optional parameter 'ground' or 'excited'

        :return: None.
        """

        if self._readoutPulse is None:
            raise Error.ArgumentError("You need to define object of class: ReadoutPulse")

        # Extract information for qubit readout simulation
        k = self.conversionLoss
        kappa = self.dissipation
        g = self.gain
        z = self.impedance
        qubitLevel = self.pulseModel.sysLevel

        # check if the qubit index exists
        if resIdx is not None:
            for idx in resIdx:
                if idx not in range(self.pulseModel.subSysNum):
                    raise Error.ArgumentError(f"Qubit index {idx} not found")
        else:
            resIdx = range(self.pulseModel.subSysNum)

        readoutPulse = self.readoutPulse

        if state == 'ground':
            psi = tensor(basis(qubitLevel, 0), basis(self.level, 0))
        elif state == 'excited':
            psi = tensor(basis(qubitLevel, 1), basis(self.level, 0))
        else:
            raise Error.ArgumentError("The input qubit state is 'ground' or 'excited'")

        # set the frequency of local oscillator for demodulation
        omegaLo = readoutPulse.loFreq

        # initialize a dict object to record the evolution
        rhoHistory = {}

        # define annihilation and creation operators
        a = tensor([identity(qubitLevel), destroy(self.level).matrix])
        adag = dagger(a)

        # define quadrature operators
        x = 0.5 * (adag + a)
        p = 1j * 0.5 * (adag - a)

        viRawList = {}
        vqRawList = {}
        viList = {}
        vqList = {}
        rhoList = {}

        # simulate the readout dynamic of the designated qubits
        for idx in resIdx:

            # run simulation of the given qubit
            ham = self._createHam(idx=idx, driveFreq=readoutPulse.driveFreq[idx])
            job = self._readoutJob(amp=readoutPulse.driveStrength[idx], duration=duration)
            res0 = ham.simulate(state0=psi, job=job, isOpen=True, recordEvolution=True)

            # Append the evolution of density matrix to idx-th qubit
            rhoHistory[idx] = res0[0]['evolution_history']

            _, maxDt = job.computeMaxTime()
            timeIdx = range(0, int(maxDt))

            # Calculate the intermediate frequency
            omegaRf = self.readoutPulse.driveFreq[idx]

            omegaIf = abs(omegaLo - omegaRf)

            viRawList[idx] = []
            vqRawList[idx] = []

            vif = k * readoutPulse.driveStrength[idx] * sqrt(kappa * g * z * omegaRf / 2)

            # Record the raw data of vi, vq quadratures for each dt
            for dtIdx in timeIdx:

                nowTime = dtIdx * self.dt
                nowRho = rhoHistory[idx][dtIdx]

                vi = vif * (x * cos(omegaIf * nowTime) - p * sin(omegaIf * nowTime))

                vq = - vif * (p * cos(omegaIf * nowTime) + x * sin(omegaIf * nowTime))

                viRawList[idx].append(trace(vi @ nowRho).real)
                vqRawList[idx].append(trace(vq @ nowRho).real)

            viList[idx] = trace(vif * x @ rhoHistory[idx][-1]).real
            vqList[idx] = trace(-vif * p @ rhoHistory[idx][-1]).real

            rhoList[idx] = rhoHistory[idx][-1]

        # Return the result in the form of dictionary data type
        readoutData = {'vi': viList,
                       'vq': vqList,
                       'viRaw': viList,
                       'vqRaw': vqList,
                       'index': resIdx,
                       'duration': duration,
                       'dt': self.dt,
                       'loFreq': omegaLo,
                       'rho': rhoList}

        return readoutData


